# lsqfitgp/tests/test_jax.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
import gvar
import jax
from jax import lax
from jax import numpy as jnp
from pytest import mark

sys.path.insert(0, '.')
from lsqfitgp import _patch_jax
import lsqfitgp as lgp
import util

rng = np.random.default_rng(202303031632)

def test_elementwise_grad_1():
    def f(x):
        return 2 * x
    x = np.arange(8.)
    with jax.checking_leaks():
        y = _patch_jax.elementwise_grad(f)(x)
    y2 = jax.vmap(jax.grad(f))(x)
    util.assert_equal(y, y2)

def test_elementwise_grad_2():
    def f(x, z):
        return 2 * x * z
    x = np.arange(8.)
    with jax.checking_leaks():
        y = _patch_jax.elementwise_grad(f, 0)(x, x)
    y2 = jax.vmap(jax.grad(f, 0))(x, x)
    util.assert_equal(y, y2)

def test_elementwise_grad_3():
    def f(x):
        return 2 * x
    x = np.arange(8.)
    with jax.checking_leaks():
        y = _patch_jax.elementwise_grad(_patch_jax.elementwise_grad(f))(x)
    y2 = jax.vmap(jax.grad(jax.grad(f)))(x)
    util.assert_equal(y, y2)

def test_elementwise_grad_4():
    def f(x, z):
        return 2 * x * z
    x = np.arange(8.)
    with jax.checking_leaks():
        y = _patch_jax.elementwise_grad(_patch_jax.elementwise_grad(f, 0), 0)(x, x)
    y2 = jax.vmap(jax.grad(jax.grad(f, 0), 0))(x, x)
    util.assert_equal(y, y2)

def test_elementwise_grad_5():
    def f(x, z):
        return 2 * x * z
    x = np.arange(8.)
    with jax.checking_leaks():
        y = _patch_jax.elementwise_grad(_patch_jax.elementwise_grad(f, 0), 1)(x, x)
    y2 = jax.vmap(jax.grad(jax.grad(f, 0), 1))(x, x)
    util.assert_equal(y, y2)

def test_elementwise_grad_6():
    def f(x, z):
        return 2 * x * z
    x = np.arange(8.)
    with jax.checking_leaks():
        y = jax.jacrev(_patch_jax.elementwise_grad(f, 0), 1)(x, x)
    y2 = jax.jacrev(jax.vmap(jax.grad(f, 0)), 1)(x, x)
    util.assert_equal(y, y2)

@mark.parametrize('maxnbytes', [1, 10 * 8, 1000 * 8])
def test_batcher(maxnbytes):
    def func(x, y):
        return x * y
    batched = _patch_jax.batchufunc(func, maxnbytes=maxnbytes)
    x, y = rng.standard_normal((2, 20, 7))
    args = (x[None, :, :], y[:, None, :])
    p1 = func(*args)
    p2 = batched(*args)
    util.assert_equal(p1, p2)

@mark.parametrize('maxnbytes', [1, 10 * 8, 1000 * 8])
def test_batcher_structured(maxnbytes):
    def func(x, y):
        return jnp.sum(x['x'] * y['x'], axis=-1)
    batched = _patch_jax.batchufunc(func, maxnbytes=maxnbytes)
    xy = rng.standard_normal((2, 20, 7)).view([('x', float, 7)]).squeeze(-1)
    xy = lgp.StructuredArray(xy)
    x, y = xy
    args = (x[None, :], y[:, None])
    p1 = func(*args)
    p2 = batched(*args)
    util.assert_allclose(p1, p2, atol=2e-15)

def test_hash():
    """ check the jax port of fast-hash against the original code """
    inputs = [
        jnp.array([], dtype=jnp.uint8),
        jnp.array([234], dtype=jnp.uint8),
        jnp.array([194, 116], dtype=jnp.uint8),
        jnp.array([160, 237, 166], dtype=jnp.uint8),
        jnp.array([72, 56, 46, 113], dtype=jnp.uint8),
        jnp.array([152, 22, 163, 7, 234], dtype=jnp.uint8),
        jnp.array([100, 11, 190, 249, 103, 74], dtype=jnp.uint8),
        jnp.array([119, 52, 46, 248, 188, 178, 216], dtype=jnp.uint8),
        jnp.array([81, 10, 197, 4, 19, 41, 69, 164], dtype=jnp.uint8),
        jnp.array([53, 246, 128, 162, 79, 228, 71, 137, 255], dtype=jnp.uint8),
        jnp.array([145, 141, 43, 100, 125, 107, 12, 4, 147, 229], dtype=jnp.uint8),
        jnp.array([117, 92, 35, 144, 76, 140, 59, 36, 42, 13, 94], dtype=jnp.uint8),
        jnp.array([91, 207, 0, 152, 226, 159, 190, 164, 136, 176, 194, 59], dtype=jnp.uint8),
        jnp.array([126, 94, 132, 168, 44, 150, 242, 165, 199, 149, 248, 82, 141], dtype=jnp.uint8),
        jnp.array([26, 101, 134, 203, 216, 141, 100, 242, 248, 225, 83, 131, 27, 100], dtype=jnp.uint8),
        jnp.array([153, 2, 211, 91, 131, 54, 101, 233, 213, 71, 216, 126, 60, 48, 157], dtype=jnp.uint8),
        jnp.array([114, 165, 8, 26, 213, 17, 112, 170, 104, 161, 164, 95, 53, 17, 149, 170], dtype=jnp.uint8),
        jnp.array([40, 198, 242, 87, 28, 55, 234, 142, 22, 200, 236, 65, 198, 91, 197, 233, 46], dtype=jnp.uint8),
        jnp.array([208, 21, 5, 101, 61, 240, 41, 134, 164, 25, 109, 253, 108, 140, 229, 255, 39, 199], dtype=jnp.uint8),
        jnp.array([240, 22, 57, 231, 226, 172, 97, 114, 34, 20, 14, 47, 118, 129, 193, 93, 43, 209, 75], dtype=jnp.uint8),
    ]
    hashes64 = jnp.array([
        7502587791032603753,
        16941272163545924368,
        10988138224395471776,
        9507901428091620561,
        8215232957141175337,
        18053746358964717198,
        12425373722766252877,
        2946925277746383721,
        12402381367179054957,
        755910146092029036,
        3255785893224143811,
        12592656301469221220,
        428602295608661196,
        5824169786726525377,
        1508071078291841094,
        11448092356368632731,
        6157277036160160880,
        9731805725958528408,
        3366289320067065534,
        17424790981778646777,
    ], dtype=jnp.uint64)
    hashes32 = jnp.array([
        1004310665,
        2046185678,
        3566082500,
        2790396102,
        2182032100,
        2323244336,
        1312080940,
        2492442272,
        1823551547,
        3569298354,
        1867203821,
        2449676296,
        2938746445,
        3041190206,
        3115372248,
        3527005061,
        1217622642,
        4177513530,
        303099792,
        2425579332,
    ], dtype=jnp.uint32)
    seed32 = 2428169863
    seed64 = 6361217807637034346
    for inp, h32, h64 in zip(inputs, hashes32, hashes64):
        hash64 = _patch_jax.fasthash64(inp, seed64)
        hash32 = _patch_jax.fasthash32(inp, seed32)
        assert h64 == hash64
        assert h32 == hash32

def genint(dtype, size=()):
    """ generate integers spanning full type range """
    return rng.integers(np.iinfo(dtype).min, np.iinfo(dtype).max, endpoint=True, dtype=dtype, size=size)

def test_hash_bitflip():
    """ check a single bit flip in the input changes 50% of the hash bits """
    buf = genint('u1', 43)
    
    bufmod = buf.copy()
    i = rng.integers(buf.size)
    j = rng.integers(8)
    bufmod[i] ^= 1 << j

    assert np.sum(lax.population_count(buf ^ bufmod)) == 1

    seed = genint('u8')
    h = _patch_jax.fasthash64(buf, seed)
    hmod = _patch_jax.fasthash64(bufmod, seed)
    diff = h ^ hmod
    flipped_bits = lax.population_count(diff)
    total_bits = h.nbytes * 8
    frac_flipped = flipped_bits / total_bits
    prob_flip = 0.5
    std = np.sqrt(prob_flip * (1 - prob_flip) / total_bits)
    assert abs(frac_flipped - prob_flip) <= 3 * std

def test_hash_numpy():
    """ check numpy arrays do not break the hash """
    buf = genint('u1', 2)
    _patch_jax.fasthash64(buf, 12345)



# lsqfitgp/examples/c.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE C.

    Where a nonlinear transformation hides the true height of some
    crosses.

"""

import lsqfitgp as lgp
import lsqfit
from matplotlib import pyplot as plt
import numpy as np
import gvar

plot_simulated_lines = True

xdata = np.linspace(0, 10, 10)
xpred = np.linspace(-15, 25, 300)

gp = lgp.GP(lgp.ExpQuad(scale=3))
gp.addx(xdata, 'data')
gp.addx(xpred, 'pred')

true_par = dict(
    phi=np.sin(xdata),
    y0=10
)

def fcn(data_or_pred, p):
    if data_or_pred == 'data':
        phi = p['phi']
    elif data_or_pred == 'pred':
        phi = gp.predfromfit({'data': p['phi']}, 'pred')
    else:
        raise KeyError(data_or_pred)
    
    return gvar.tanh(1 + phi) + p['y0']

yerr = 0.05
ysdev = yerr * np.ones(len(xdata))
true_y = fcn('data', true_par)
ymean = true_y + ysdev * np.random.randn(len(ysdev))
y = gvar.gvar(ymean, ysdev)

prior = dict(
    phi=gp.prior('data'),
    y0=gvar.gvar(0, 1000)
)

p0=dict(
    phi=np.random.multivariate_normal(np.zeros(len(xdata)), gvar.evalcov(prior['phi']))
)

fit = lsqfit.nonlinear_fit(data=('data', y), prior=prior, fcn=fcn, p0=p0)
print(fit.format(maxline=True))

ypred = fcn('pred', fit.p)
ypredalt = fcn('pred', fit.palt)

phipred = gp.predfromfit({'data': fit.p['phi']}, 'pred')

fig, axs = plt.subplots(1, 2, num='c', clear=True)

for ax, variable in zip(axs, ['y', 'phi']):
    ax.set_title(variable)
    
    for label in 'pred', 'predalt':
        if variable == 'phi' and label == 'predalt':
            continue
        
        pred = eval(variable + label)

        m = gvar.mean(pred)
        s = gvar.sdev(pred)

        patch = ax.fill_between(xpred, m - s, m + s, label=label, alpha=0.5)
        color = patch.get_facecolor()[0]
    
        if plot_simulated_lines:
            cov = gvar.evalcov(pred)
            simulated_lines = np.random.multivariate_normal(m, cov, size=10)
            ax.plot(xpred, simulated_lines.T, '-', color=color)
    
    if variable == 'phi':
        ax.plot(xdata, true_par['phi'], 'rx', label='true')

axs[0].errorbar(xdata, gvar.mean(y), yerr=gvar.sdev(y), fmt='k.', label='data')
axs[0].plot(xdata, true_y, 'rx', label='true')

for ax in axs:
    ax.legend(loc='best')

fig.tight_layout()
fig.show()



from scipy import special
import numpy as np
from matplotlib import pyplot as plt

fig, ax = plt.subplots(num='bessel2', clear=True)

nu = np.linspace(-10, 0, 10)
x = np.linspace(0, 5, 100)

for v in nu:
    y = (x / 2) ** -v * special.jv(v, x) * special.gamma(v + 1)
    ax.plot(x, y, label=f'$\\nu$={v:.2g}')

ax.legend()
ax.set_xlabel('$x$')
ax.set_ylabel('$J_\\nu(x)$')
ax.grid()

fig.show()



"""Fit of parton distributions functions (PDFs)

Like pdf6, but with hyperparameters"""

import lsqfitgp as lgp
import numpy as np
from scipy import linalg
from matplotlib import pyplot as plt
import gvar
import lsqfit
import time

np.random.seed(20220416)

#### SETTINGS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

indices = dict(
    # quark, antiquark
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

pid  = flavor['f0']
name = flavor['f1']

nflav  = len(flavor)

# linear data
nx        = 30 # number of PDF points used for the transformation
ndata     = 10 # number of datapoints
rankmcov  =  9 # rank of the covariance matrix of the theory error

# quadratic data
nx2       = 30 # must be <= nx
ndata2    = 10
rankmcov2 =  9

#### MODEL ####
# for each PDF:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

hyperprior = {
    'log(scale)': np.log(gvar.gvar(0.5, 0.5)),
}
def makekernel(hp):
    return lgp.ExpQuad(dim='x', scale=hp['scale']) * lgp.White(dim='pid')

# grid of points to which we apply the transformation
xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

# linear map PDF(xdata) -> data
Mcomps = np.random.randn(rankmcov, ndata, nflav, nx)
Mcomps /= np.sqrt(Mcomps.size / ndata)
Mparams = gvar.gvar(np.random.randn(rankmcov), np.full(rankmcov, 0.1))
M = lambda params: np.tensordot(params, Mcomps, 1)

# quadratic map PDF(xdata) -> data2
M2comps = np.random.randn(rankmcov2, ndata2, nflav, nx2, nx2)
M2comps /= 2 * np.sqrt(M2comps.size / ndata2)
M2comps = (M2comps + np.swapaxes(M2comps, -1, -2)) / 2
M2params = gvar.gvar(np.random.randn(rankmcov2), np.full(rankmcov2, 0.1))
M2 = lambda params: np.tensordot(params, M2comps, 1)

# endpoints of the integral for each PDF
xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

# matrix to subtract the endpoints
suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

# matrix to subtract two quarks
qdiff = np.array([1, -1])[:, None]

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### GP OBJECT ####

def makegp(hp):
    gp = lgp.GP()

    kernel = makekernel(hp)
    
    gp.addproc(kernel, 'h')
    gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
    gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
    gp.addproctransf({
        'primitive': lambda x: x['x'],
        'h'        : -1,
    }, 'primitive of xf(x)')

    gp.addx(xdata, 'xdata', proc='f')

    # linear data (used for warmup fit)
    gp.addtransf({'xdata': M(gvar.mean(Mparams))}, 'data', axes=2)

    # total momentum rule
    gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
    gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)

    # quark sum rules
    for quark in 'ducs':
        idx = indices[quark] # [quark, antiquark] indices
        label = f'{quark}{quark}bar' # the one appearing in `constraints`
        xlabel = f'x{label}'
        gp.addx(xinteg[idx], xlabel, proc='primitive')
        gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)
    
    return gp
    
#### NONLINEAR FUNCTION ####

def fcn(params):
    xdata = params['xdata']
    Mparams = params['Mparams']
    M2params = params['M2params']
    
    data = np.tensordot(M(Mparams), xdata, 2)
    
    # data2 = np.einsum('dfxy,fx,fy->d', M2, xdata, xdata)
    # np.einsum does not work with gvar
    xdata2 = xdata[:, None, :nx2] * xdata[:, :nx2, None]
    data2 = np.tensordot(M2(M2params), xdata2, 3)
    
    return dict(data=data, data2=data2)

def makeprior(gp):
    prior = gp.predfromdata(constraints, ['xdata'])
    prior['Mparams'] = Mparams
    prior['M2params'] = M2params
    return prior

#### FAKE DATA ####

truehp = gvar.sample(hyperprior)
truegp = makegp(truehp)
trueparams = gvar.sample(makeprior(truegp))
truedata = fcn(trueparams)

dataerr = {
    k: np.full_like(v, 0.1 * (np.max(v) - np.min(v)))
    for k, v in truedata.items()
}
data = gvar.make_fake_data(gvar.gvar(truedata, dataerr))

# check sum rules approximately with trapezoid rule
def check_integrals(x, y):
    checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
    print('sum_i int dx x f_i(x) =', checksum)
    for q in 'ducs':
        idx = indices[q]
        qx = x[idx]
        qy = y[idx]
        checksum = np.sum(qdiff * (qy[:, 1:] + qy[:, :-1]) / 2 * np.diff(qx, axis=1))
        print(f'sum_i={q}{q}bar int dx f_i(x) =', checksum)

print('check integrals in fake data:')
check_integrals(xdata['x'], trueparams['xdata'])

#### FIT ####

# find the minimization starting point with a simplified version of the fit
easyfit = truegp.predfromdata(dict(data=data['data'], **constraints), ['xdata'])
p0 = gvar.mean(easyfit)

hyperprior = gvar.BufferDict(hyperprior)
hypermean = gvar.mean(hyperprior.buf)
hypercov = gvar.evalcov(hyperprior.buf)
hyperchol = linalg.cholesky(hypercov)

i = 0
lasttime = time.time()
def analyzer(hp):
    global i, lasttime
    i += 1
    now = time.time()
    interval = now - lasttime
    print(f'iteration {i:3d} ({interval:#.2g} s): {hp}')
    lasttime = now

def fitargs(hp):
    analyzer(hp)
    gp = makegp(hp)
    prior = makeprior(gp)
    args = dict(
        data = data,
        fcn = fcn,
        prior = prior,
        p0 = p0,
    )
    residuals = linalg.solve_triangular(hyperchol, hp.buf - hypermean)
    plausibility = -1/2 * (residuals @ residuals)
    return args, plausibility

z0 = gvar.mean(hyperprior)
fit, fithp = lsqfit.empbayes_fit(z0, fitargs)
print(fit.format(maxline=True, pstyle='v'))
print(fit.format(maxline=-1))

gp = makegp(fithp)
prior = makeprior(gp)
pred = fcn(fit.p)

print('check integrals in fit:')
check_integrals(xdata['x'], fit.p['xdata'])

print('\nhyperparameters (true, fitted):')
for k in fithp.all_keys():
    print(f'{k}: {truehp[k]:#.2g}\t{fithp[k]:#.2g}')

#### PLOT RESULTS ####

fig, axs = plt.subplots(2, 3, num='pdf7', clear=True, figsize=[13, 8])
axs[0, 0].set_title('PDFs')
axs[1, 0].set_title('PDFs')
axs[0, 1].set_title('Data')
axs[1, 1].set_title('Data (quadratic)')
axs[0, 2].set_title('M parameters')
axs[1, 2].set_title('M2 parameters')

ax = axs[0, 0]

for i in range(nflav):
    
    if i >= 4:
        ax = axs[1, 0]
    
    x = xdata[i]['x']
    ypdf = fit.p['xdata'][i]
    ydata = trueparams['xdata'][i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    
    color = 'C' + str(i // 2)
    if i % 2:
        kw = dict(hatch='//////', edgecolor=color, facecolor='none')
        kwp = dict(color=color, linestyle='--')
    else:
        kw = dict(alpha=0.6, facecolor=color)
        kwp = dict(color=color)
    
    ax.fill_between(x, m - s, m + s, label=name[i], **kw)
    ax.plot(x, ydata, **kwp)

for ax in axs[:, 0]:
    ax.legend(fontsize='small')

for ax, label in zip(axs[:, 1], ['data', 'data2']):
    d = pred[label]
    m = gvar.mean(d)
    s = gvar.sdev(d)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    d = data[label]
    ax.errorbar(x, gvar.mean(d), gvar.sdev(d), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, truedata[label], drawstyle='steps-mid', color='black', label='truth')

for ax, label in zip(axs[:, 2], ['Mparams', 'M2params']):
    p = fit.p[label]
    m = gvar.mean(p)
    s = gvar.sdev(p)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    p = prior[label]
    ax.errorbar(x, gvar.mean(p), gvar.sdev(p), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, trueparams[label], drawstyle='steps-mid', color='black', label='truth')

for ax in axs[:, 1:].flat:
    ax.legend()

fig.tight_layout()
fig.show()



"""Autograd makes array.reshape behave like np.reshape(array) which has a
different signature, but only in a certain corner case I have not properly
figured out."""

import autograd
from autograd import numpy as anp

x = anp.arange(9)

def f(x):
    return (x ** 2).reshape(3, 3)

jac = autograd.jacobian(f)
hess = autograd.jacobian(jac)

J = jac(x)
H = hess(x)



class cippa:
    
    def __init_subclass__(cls):
        cls.cippa = 'original cippa'
    
class lippa:
    
    def __init_subclass__(cls):
        cls.cippa = 'fake cippa'

class fagiolino(cippa, lippa):
    
    def __init_subclass__(cls):
        pass


import lsqfitgp as lgp
import jax
import numpy as np

@jax.jit
def f():
	dec = lgp._linalg.Chol(np.array([[2]]))
	print('@@@@@@@@@@@@@@@@@@@', id(dec), '@@@@@@@@@@@@@@@@@@@')
	return dec.logdet()

with jax.checking_leaks():
	print(f())



# lsqfitgp/tests/test_jax.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# TODO maybe use sympy as reference, since it's better maintained than mpmath

import sys

import jax
from jax import test_util
import numpy as np
from scipy import special, linalg
import pytest
from pytest import mark
import mpmath

sys.path.insert(0, '.')
from lsqfitgp import _special, _patch_jax
import util

gen = np.random.default_rng(202207201419)

def test_sinc():
    x = np.linspace(-0.1, 0.1, 1000)
    s1 = np.sinc(x)
    s2 = _special.sinc(x)
    util.assert_allclose(s2, s1, atol=1e-15, rtol=1e-15)

def test_jvmodx2():
    nu = np.linspace(-5, 5, 20)
    x = np.linspace(1e-15, 0.1, 1000)
    for v in nu:
        s1 = (x / 2) ** -v * special.jv(v, x)
        s2 = _special.jvmodx2(v, x ** 2)
        util.assert_allclose(s2, s1, atol=1e-15, rtol=1e-14)
        util.assert_allclose(_special.jvmodx2(v, 0), 1 / special.gamma(v + 1), rtol=1e-14)
        test_util.check_grads(lambda x: _special.jvmodx2(v, x ** 2), (x,), 2)

def test_kvmodx2():
    nu = np.linspace(-5, 5, 20)
    x = np.linspace(1e-15, 0.1, 1000)
    xsoft = np.linspace(1, 10, 1000)
    for v in nu:
        s1 = 2 / special.gamma(v) * (x / 2) ** v * special.kv(v, x)
        s2 = _special.kvmodx2(v, x ** 2)
        util.assert_allclose(s2, s1, atol=1e-15, rtol=1e-14)
        util.assert_allclose(_special.kvmodx2(v, 0), 1, rtol=1e-14)
        test_util.check_grads(lambda x: _special.kvmodx2(v, x ** 2), (xsoft,), 2)
        if v >= 0.5: # negative diverges, and below 0.5 d/dx at 0 is inf
            for no in range(5):
                util.assert_allclose(_special.kvmodx2(v, 1e-15, no), _special.kvmodx2(v, 0, no), rtol=1e-11)
        # TODO really need to check at 0 in other cases
    xz = np.linspace(0, 1, 1000)
    np.testing.assert_equal(_special.kvmodx2(0, xz), np.where(xz, 0, 1))

def test_kvmodx2_hi():
    x = np.linspace(1e-15, 10, 1000)
    x2 = x ** 2
    for p in range(5):
        v = p + 1/2
        f1 = lambda x: _special.kvmodx2_hi(x, p)
        f2 = lambda x: _special.kvmodx2(v, x)
        for _ in range(3):
            s1 = f1(x2)
            s2 = f2(x2)
            util.assert_allclose(s1, s2, atol=0, rtol=1e-14)
            f1 = _patch_jax.elementwise_grad(f1)
            f2 = _patch_jax.elementwise_grad(f2)

def randpoly(n):
    while True:
        a = gen.random.standard_normal(n)
        if a[0] != 0:
            return a

def test_exp1_imag_and_ci():
    xs = [
        np.linspace(1e-15, 100, 1000),
        np.logspace(2, 20, 1000),
    ]
    for x in xs:
        y1 = _special.exp1_imag(x)
        y2 = special.exp1(-1j * x)
        util.assert_allclose(y1, y2, atol=0, rtol=1e-14)
        
        y1 = _special.ci(x)
        _, y2 = special.sici(x)
        util.assert_allclose(y1, y2, atol=1e-15, rtol=1e-15)

def test_expm1x():
    x = np.linspace(-2, 2, 10000)
    y = _special.expm1x(x)
    y2 = x * x / 2 * special.hyp1f1(1, 3, x)
    np.testing.assert_array_max_ulp(y, y2, 11)
    y = _special.expm1x(x.astype('f'))
    np.testing.assert_array_max_ulp(y, y2.astype('f'), 4)
    test_util.check_grads(_special.expm1x, (x,), 2)

zeta = np.vectorize(lambda *args: float(mpmath.zeta(*args)))

def test_hurwitz_zeta():
    s = np.linspace(-10, 0, 100)[:, None]
    a = np.linspace(0, 1, 100)
    z1 = zeta(s, a)
    z2 = _special.hurwitz_zeta(s, a)
    eps = np.finfo(float).eps
    tol = 350 * eps * np.max(np.abs(z1), 1)
    maxdiff = np.max(np.abs(z2 - z1), 1)
    assert np.all(maxdiff < tol)

def test_hurwitz_zeta_vectorized():
    s = np.linspace(-10, 0, 100)
    a = np.linspace(0, 1, 100)
    z1 = _special.hurwitz_zeta(s, a)
    z2 = np.vectorize(_special.hurwitz_zeta)(s, a)
    np.testing.assert_array_max_ulp(z1, z2, 780)
    # TODO what?? 675 ULP?? what?? => 780 with latest ubuntu release!

def test_gamma():
    x = np.linspace(-100, 100, 1000)
    g1 = special.gamma(x)
    g2 = _special.gamma(x)
    np.testing.assert_array_max_ulp(g2, g1, 1500)

def _periodic_zeta_mpmath(x, s):
    with mpmath.workdps(32):
        arg = mpmath.exp(2j * mpmath.pi * x)
        return complex(mpmath.polylog(s, arg))
        # the doubled working precision serves two purposes:
        # 1) compute accurately arg, for example, with x=1 the result would
        #    have a small imaginary part which changes the result significantly
        # 2) polylog is inaccurate for s near an integer (mpmath issue #634)

def _zeta_mpmath(s):
    return complex(mpmath.zeta(s))

@np.vectorize
def periodic_zeta(x, s):
    if int(x) == x:
        return _zeta_mpmath(s)
        # patch for mpmath.polylog(s, 1) != zeta(s)
    else:
        return _periodic_zeta_mpmath(x, s)

@mark.parametrize('i', [
    pytest.param(False, id='real'),
    pytest.param(True, id='imag'),
])
@mark.parametrize('sgn', [
    pytest.param(1, id='pos'),
    pytest.param(-1, id='neg'),
])
@mark.parametrize('d', [
    pytest.param(0, id='at'),
    pytest.param(1e-4, id='close'),
    pytest.param(1e-13, id='veryclose'),
])
@mark.parametrize('s', [
    pytest.param(1 + 1e-15, id='near1'),
    pytest.param(0.5 + np.arange(1, 15), id='half'),
    pytest.param(np.arange(2, 15, 2), id='even'),
    pytest.param(np.arange(3, 15, 2), id='odd'),
])
def test_periodic_zeta(s, d, sgn, i):
    if d == 0 and sgn < 0:
        pytest.skip()

    x = np.linspace(-1, 2, 52)
    s = np.atleast_1d(s)[:, None] + sgn * d
    
    if np.any(s <= 1):
        pytest.skip()
    
    z1 = periodic_zeta(x, s)
    z1 = z1.imag if i else z1.real
    z2 = _special.periodic_zeta(x, s, i)
    
    eps = np.finfo(float).eps
    tol = 110 * eps * np.max(np.abs(z1), 1)
    maxdiff = np.max(np.abs(z2 - z1), 1)
    assert np.all(maxdiff < tol)

@mark.parametrize('i', [
    pytest.param(False, id='real'),
    pytest.param(True, id='imag'),
])
def test_periodic_zeta_deriv(i):
    x = np.linspace(-1, 2, 52)
    s = np.linspace(2.01, 16, 20)[:, None]

    z1 = 2j * np.pi * periodic_zeta(x, s - 1)
    z1 = z1.imag if i else z1.real
    z2 = _patch_jax.elementwise_grad(_special.periodic_zeta)(x, s, i)
    
    eps = np.finfo(float).eps
    tol = 100 * eps * np.max(np.abs(z1), 1)
    maxdiff = np.max(np.abs(z2 - z1), 1)
    assert np.all(maxdiff < tol)

@mark.parametrize('s', [
    pytest.param((1 - 1e-15) * np.linspace(-1, 1, 101), id='widerange'),
    pytest.param(0.5 * np.linspace(-1, 1, 101), id='medrange'),
    pytest.param(1e-8 * np.linspace(-1, 1, 101), id='shortrange'),
    pytest.param(1e-14 * np.linspace(-1, 1, 101), id='tinyrange'),
])
def test_zeta_zero(s):
    with mpmath.workdps(40):
        z1 = np.array([float(mpmath.zeta(s) - mpmath.zeta(0)) for s in s])
    z2 = _special.zeta_zero(s)
    np.testing.assert_array_max_ulp(z1, z2, 2)

@mark.parametrize('s', [
    pytest.param(1, id='pos'),
    pytest.param(-1, id='neg'),
])
@mark.parametrize('e', [
    pytest.param(0.5 * np.linspace(0, 1, 51), id='medrange'),
    pytest.param(1e-8 * np.linspace(0, 1, 51), id='shortrange'),
    pytest.param(1e-14 * np.linspace(0, 1, 51), id='tinyrange'),
])
@mark.parametrize('x', [
    pytest.param(2, id='2'),
    pytest.param(np.arange(3, 15), id='farpole'),
])
def test_gamma_incr(x, e, s):
    x = np.reshape(x, (-1, 1))
    @np.vectorize
    def func(x, e):
        with mpmath.workdps(40):
            x = mpmath.mpf(float(x))
            e = mpmath.mpf(float(e))
            denom = mpmath.gamma(x) * mpmath.gamma(1 + e)
            return float(mpmath.gamma(x + e) / denom - 1)
    e = e * s
    e = np.where(x == 1, np.abs(e), e)
    g1 = func(x, e)
    g2 = _special.gamma_incr(x, e)
    np.testing.assert_array_max_ulp(g2, g1, 7)

@mark.parametrize('s', [
    pytest.param(1, id='pos'),
    pytest.param(-1, id='neg'),
])
@mark.parametrize('x', [
    pytest.param(0.5 * np.linspace(0, 1, 51), id='medrange'),
    pytest.param(1e-8 * np.linspace(0, 1, 51), id='shortrange'),
    pytest.param(1e-14 * np.linspace(0, 1, 51), id='tinyrange'),
])
def test_gammaln1(x, s):
    @np.vectorize
    def func(x):
        with mpmath.workdps(40):
            x = mpmath.mpf(float(x))
            return float(mpmath.loggamma(1 + x))
    x = x * s
    g1 = func(x)
    g2 = _special.gammaln1(x)
    np.testing.assert_array_max_ulp(g2, g1, 2)

@mark.parametrize('s', [
    pytest.param(1, id='pos'),
    pytest.param(-1, id='neg'),
])
@mark.parametrize('a', [
    pytest.param(0.5 * np.linspace(0.04, 1, 25), id='medrange'),
    pytest.param(1e-8 * np.linspace(0.04, 1, 25), id='shortrange'),
    pytest.param(1e-14 * np.linspace(0.04, 1, 25), id='tinyrange'),
])
@mark.parametrize('x', [
    pytest.param(0, id='xzero'),
    pytest.param(np.logspace(-100, -3, 25), id='xnearzero'),
    pytest.param(np.linspace(0.02, 0.5, 25), id='xother'),
])
@mark.parametrize('q', [
    pytest.param(0, id='qzero'),
    pytest.param(1, id='qone'),
    pytest.param(np.arange(2, 11, 2), id='qeven'),
    pytest.param(np.arange(3, 11, 2), id='qodd'),
])
def test_power_diff(x, q, a, s):
    if np.any(q == 0) and np.any(s > 0):
        pytest.skip()
    x = np.reshape(x, (-1, 1, 1))
    q = np.reshape(q, (-1, 1))
    a = a * s # don't use inplace *= since arguments are not copied
    @np.vectorize
    def func(x, q, a):
        with mpmath.workdps(40):
            x = mpmath.mpf(float(x))
            q = mpmath.mpf(float(q))
            a = mpmath.mpf(float(a))
            num = mpmath.gamma(1 + q - a)
            denom = mpmath.gamma(1 - a) * mpmath.gamma(1 + q)
            zeta = mpmath.zeta(a)
            term = 2 * num / denom * zeta * x ** q
            power = x ** (q - a)
            return float(power + term)
    p1 = func(x, q, a)
    p2 = _special.zeta_series_power_diff(x, q, a)
    if np.all(q <= 1) and np.any(x):
        tol = 20 * np.max(np.abs(p1), 0) * np.finfo(float).eps
        maxdiff = np.max(np.abs(p1 - p2), 0)
        assert np.all(maxdiff < tol)
    else:
        tol = np.finfo(float).eps ** 2 / 2
        cond = np.abs(p1) < tol
        cp1 = np.where(cond, tol, p1)
        cp2 = np.where(cond, tol, p2)
        np.testing.assert_array_max_ulp(cp2, cp1, 22)

@mark.parametrize('s', [
    pytest.param(-2, id='-2'),
    pytest.param(-np.arange(4, 11, 2), id='small'),
    pytest.param(-np.arange(12, 101, 2), id='large'),
])
def test_zeta_zeros(s):
    @np.vectorize
    def func(s):
        with mpmath.workdps(20):
            return float(mpmath.diff(mpmath.zeta, s))
    def handwritten(s):
        pi = 2 * (2 * np.pi) ** (s - 1)
        n = np.around(s)
        sgn = np.where(n % 4, -1, 1)
        cos = np.pi / 2 * sgn
        gamma = _special.gamma(1 - s)
        zeta = _special.zeta(1 - s)
        return pi * cos * gamma * zeta
    z0 = handwritten(s)
    z1 = func(s)
    ulp = 25 if np.all(s >= -10) else 1113
    np.testing.assert_array_max_ulp(z0, z1, ulp)
    eps = 1e-30
    z2 = _special.zeta(eps, s) / eps
    np.testing.assert_array_max_ulp(z2, z1, ulp)

def test_zeta():
    n = np.arange(-10, 60)[:, None]
    s = np.linspace(-0.5, 0.5, 101)
    @np.vectorize
    def func(s, n):
        with mpmath.workdps(32):
            n = mpmath.mpmathify(n)
            s = mpmath.mpmathify(s)
            return float(mpmath.zeta(s + n)) if n + s != 1 else np.inf
    z1 = func(s, n)
    z2 = _special.zeta(s, n)
    np.testing.assert_array_max_ulp(z2, z1, 72)

def bernoulli_poly_handwritten(n, x):
    return [
        lambda x: 1,
        lambda x: x - 1/2,
        lambda x: x**2 - x + 1/6,
        lambda x: x**3 - 3/2 * x**2 + 1/2 * x,
        lambda x: x**4 - 2 * x**3 + x**2 - 1/30,
        lambda x: x**5 - 5/2 * x**4 + 5/3 * x**3 - 1/6 * x,
        lambda x: x**6 - 3 * x**5 + 5/2 * x**4 - 1/2 * x**2 + 1/42
    ][n](x)

def check_bernoulli(n, x):
    r1 = bernoulli_poly_handwritten(n, x)
    r2 = _special.periodic_bernoulli(n, x)
    util.assert_allclose(r1, r2, atol=1e-15, rtol=1e-9)

def test_bernoulli():
    for n in range(7):
        x = gen.uniform(0, 1, size=100)
        check_bernoulli(n, x)

@np.vectorize
def expint(n, z):
    return complex(mpmath.expint(n, z))

def test_expn():
    x = gen.uniform(0, 30, 10)
    n = np.arange(2, 12)
    result_64 = np.array([_special.expn_imag(n, x) for n in n])
    result_32 = np.array([_special.expn_imag(n.astype('i4'), x.astype('f')) for n in n])
    assert result_64.dtype == 'complex128'
    assert result_32.dtype == 'complex64'
    sol = expint(n[:, None], -1j * x)
    util.assert_allclose(result_64, sol, rtol=1e-7) # TODO quite bad
    util.assert_allclose(result_32, sol, atol=0.01) # TODO very bad!!

def test_kvp():
    test_util.check_grads(lambda z: _special.kv(3.2, z), (1.5,), 2)

# TODO these tests currently take 4 min out of 17 total. I guess the bottleneck
# is mpmath. I should produce and commit a cache of values.



# lsqfitgp/_GP.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import functools
import sys
import abc
import warnings

import gvar
import numpy
import jax
from jax import numpy as jnp
from jax.scipy import linalg as jlinalg
from scipy import linalg, sparse

from . import _Kernel
from . import _linalg
from . import _array
from . import _Deriv
from . import _patch_jax
from . import _patch_gvar

__all__ = [
    'GP',
]

# TODO make many methods that do not return anything return self, to allow
# a polars-like syntax as lgp.GP(kernel).addx(x, 0).addx(x, 1).prior() => make
# GP immutable and return a new GP, with non-copies of actual data? Kernels, jax
# arrays and StructuredArray are already immutable.

# TODO change the method names that define processes to def* instead of add*,
# such that the distinction between the two kinds is clear, and take the
# occasion to break the parameter order: make a coherent standard
# name of new thing, operation, name of old thing, *, options
#
# defproc(key, kernel, *, deriv=0)
# defproctransf(key, ops, *, deriv=0)
# defproclintransf(key, transf, procs, *, deriv=0, checklin=False)
# defprocderiv(key, deriv, proc)
# defprocxtransf(key, transf, proc)
# defprocrescale(key, scalefun, proc)
#
# Do not use DefaultProcess implicitly, document it as GP.DefaultProcess that
# is used as default by instatiation methods

class GP:
    """
    
    Object that represents a Gaussian process over arbitrary input.
    
    Methods
    -------
    addx
        Add points where the Gaussian process is evaluated.
    addlintransf
        Define a finite linear transformation of the evaluated process.
    addtransf
        Define a finite linear transformation of the evaluated process with
        explicit coefficients.
    addcov
        Introduce a set of user-provided prior covariance matrix blocks.
    addproc
        Define a new independent component of the process.
    addproctransf
        Define a pointwise linear transformation of the process with explicit
        coefficients.
    addproclintransf
        Define a pointwise linear transformation of the process.
    addkernelop
        Define a transformation of the process through a kernel method.
    addprocderiv
        Define a derivative of the process.
    addprocxtransf
        Define a process with transformed inputs.
    addprocrescale
        Define a rescaled process.
    prior
        Compute the prior for the process.
    pred
        Compute the posterior for the process.
    predfromfit
        Like `pred` with ``fromdata=False``.
    predfromdata
        Like `pred` with ``fromdata=True``.
    marginal_likelihood
        Compute the marginal likelihood, i.e., the unconditional probability of
        data.
    decompose
        Decompose a pos. def. matrix.
    
    Parameters
    ----------
    covfun : Kernel or None
        An instance of `Kernel` representing the covariance kernel of the
        default process of the GP object. It can be left unspecified.
    solver : str
        The algorithm used to decompose the prior covariance matrix. See
        :meth:`~GP.decompose` for the available solvers. Default is `eigcut+`.
    checkpos : bool
        If True (default), raise a `LinAlgError` if the prior covariance matrix
        turns out non positive within numerical error.
    checksym : bool
        If True (default), check that the prior covariance matrix is
        symmetric. If False, only half of the matrix is computed.
    checkfinite : bool
        If True (default), check that the prior covariance matrix does not
        contain infs or nans.
    checklin : bool
        If True (default), the method `addlintransf` will check that the
        given transformation is linear on a random input tensor.
    posepsfac : number
        The threshold used to check if the prior covariance matrix is positive
        definite is multiplied by this factor (default 1).
    halfmatrix : bool
        If True and `checksym=False`, compute only half of the covariance
        matrices by unrolling their lower triangular part as flat arrays. This
        may actually be a large performance hit if the input arrays have large
        item size or if the implementation of the kernel takes advantage of
        non-broadcasted inputs, so it is False by default.
    **kw
        Additional keyword arguments are passed to the solver, see
        :meth:`~GP.decompose`.

    """
    
    @staticmethod
    def _concatenate(alist):
        """
        Decides to use numpy.concatenate or jnp.concatenate depending on the
        input to support gvars.
        """
        if any(a.dtype == object for a in alist):
            return numpy.concatenate(alist)
        else:
            return jnp.concatenate(alist)

    @staticmethod
    def _triu_indices_and_back(n):
        """
        Return indices to get the upper triangular part of a matrix, and indices
        to convert a flat array of upper triangular elements to a symmetric
        matrix.
        """
        ix, iy = jnp.triu_indices(n)
        q = jnp.empty((n, n), ix.dtype)
        a = jnp.arange(ix.size)
        q = q.at[ix, iy].set(a)
        q = q.at[iy, ix].set(a)
        return ix, iy, q

    @staticmethod
    def _compatible_dtypes(d1, d2): # pragma: no cover
        """
        Function to check x arrays datatypes passed to GP.addx. If the dtype is
        structured, it checks the structure of the fields is the same, but
        allows casting of concrete dtypes (like, in one array a field can be
        int, in another float, as long as the field name and position is the
        same). Currently not used.

        May not be needed in numpy 1.23, check what result_type does now.
        """
        if d1.names != d2.names or d1.shape != d2.shape:
            return False
        if d1.names is not None:
            for name in d1.names:
                if not _compatible_dtypes(d1.fields[name][0], d2.fields[name][0]):
                    return False
        else:
            try:
                numpy.result_type(d1, d2) # TODO not strict enough!
            except TypeError:
                return False
        return True
    
    class _Element(metaclass=abc.ABCMeta):
        """
        Abstract class for an object holding information associated to a key in
        a GP object.
        """
    
        @property
        @abc.abstractmethod
        def shape(self): # pragma: no cover
            """Output shape"""
            pass
    
        @property
        def size(self):
            return numpy.prod(self.shape, dtype=int)

    class _Points(_Element):
        """Points where the process is evaluated"""
    
        def __init__(self, x, deriv, proc):
            assert isinstance(x, (numpy.ndarray, jnp.ndarray, _array.StructuredArray))
            assert isinstance(deriv, _Deriv.Deriv)
            self.x = x
            self.deriv = deriv
            self.proc = proc
    
        @property
        def shape(self):
            return self.x.shape

    class _LinTransf(_Element):
        """Linear transformation of other _Element objects"""
    
        shape = None
    
        def __init__(self, transf, keys, shape):
            self.transf = transf
            self.keys = keys
            self.shape = shape
    
        def matrices(self, gp):
            """
            Matrix coefficients of the transformation (with flattened inputs
            and output)
            """
            elems = [gp._elements[key] for key in self.keys]
            matrices = []
            transf = jax.vmap(self.transf, 0, 0)
            for i, elem in enumerate(elems):
                inputs = [
                    jnp.eye(elem.size).reshape((elem.size,) + elem.shape)
                    if j == i else
                    jnp.zeros((elem.size,) + ej.shape)
                    for j, ej in enumerate(elems)
                ]
                output = transf(*inputs).reshape(elem.size, self.size).T
                matrices.append(output)
            return matrices

    class _Cov(_Element):
        """User-provided covariance matrix block(s)"""
    
        shape = None
    
        def __init__(self, blocks, shape):
            """ blocks = dict (key, key) -> matrix """
            self.blocks = blocks
            self.shape = shape

    class _Proc(metaclass=abc.ABCMeta):
        """
        Abstract base class for an object holding information about a process
        in a GP object.
        """
    
        @abc.abstractmethod
        def __init__(self): # pragma: no cover
            pass
    
    class _ProcKernel(_Proc):
        """An independent process defined with a kernel"""
    
        def __init__(self, kernel, deriv=0):
            assert isinstance(kernel, _Kernel.Kernel)
            self.kernel = kernel
            self.deriv = deriv
    
    class _ProcTransf(_Proc):
        """A process defined as a linear transformation of other processes"""
    
        def __init__(self, ops, deriv):
            """ops = dict proc key -> callable"""
            self.ops = ops
            self.deriv = deriv

    class _ProcLinTransf(_Proc):
    
        def __init__(self, transf, keys, deriv):
            self.transf = transf
            self.keys = keys
            self.deriv = deriv

    class _ProcKernelOp(_Proc):
        """A process defined by an operation on the kernel of another process"""
    
        def __init__(self, proc, method, arg):
            """proc = proc key, method = Kernel method, arg = argument to method"""
            self.proc = proc
            self.method = method
            self.arg = arg
        
    class _SingletonMeta(type):
    
        def __repr__(cls):
            return cls.__name__

    class _Singleton(metaclass=_SingletonMeta):
    
        def __new__(cls):
            raise NotImplementedError(f"{cls.__name__} can not be instantiated")

    class DefaultProcess(_Singleton):
        """ Key for the default process in GP objects """
        pass

    _zerokernel = _Kernel.Zero()

    def __init__(self,
        covfun=None,
        *,
        solver='eigcut+',
        checkpos=True,
        checksym=True,
        checkfinite=True,
        checklin=True,
        posepsfac=1,
        halfmatrix=False,
        **kw,
    ):
        self._elements = dict() # key -> _Element
        self._covblocks = dict() # (key, key) -> matrix
        self._priordict = gvar.BufferDict({}, dtype=object) # key -> gvar array (shaped)
        self._decompcache = dict() # tuple of keys -> Decomposition
        self._procs = dict() # proc key -> _Proc
        self._kernels = dict() # (proc key, proc key) -> _KernelBase
        if covfun is not None:
            if not isinstance(covfun, _Kernel.Kernel):
                raise TypeError('covariance function must be of class Kernel')
            self._procs[self.DefaultProcess] = self._ProcKernel(covfun)
        decomp = self._getdecomp(solver)
        self._decompclass = lambda K, **kwargs: decomp(K, **kwargs, **kw)
        self._checkpositive = bool(checkpos)
        self._posepsfac = float(posepsfac)
        self._checksym = bool(checksym)
        self._checkfinite = bool(checkfinite)
        self._checklin = bool(checklin)
        self._decompname = str(solver)
        self._halfmatrix = bool(halfmatrix)
        assert not (halfmatrix and checksym)

    def addproc(self, kernel=None, key=DefaultProcess, *, deriv=0):
        """
        
        Add an independent process.
        
        Parameters
        ----------
        kernel : Kernel
            A kernel for the process. If None, use the default kernel. The
            difference between the default process and a process defined with
            the default kernel is that, although they have the same kernel,
            they are independent.
        key : hashable
            The name that identifies the process in the GP object. If not
            specified, sets the kernel of the default process.
        deriv : Deriv-like
            Derivatives to take on the process defined by the kernel.
                
        """
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        if kernel is None:
            kernel = self._procs[self.DefaultProcess].kernel
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = self._ProcKernel(kernel, deriv)
    
    def addproctransf(self, ops, key=DefaultProcess, *, deriv=0):
        """
        
        Define a new process as a linear combination of other processes.
        
        Let f_i(x), i = 1, 2, ... be already defined processes, and g_i(x) be
        deterministic functions. The new process is defined as
        
            h(x) = g_1(x) f_1(x) + g_2(x) f_2(x) + ...
        
        Parameters
        ----------
        ops : dict
            A dictionary mapping process keys to scalars or scalar
            functions. The functions must take an argument of the same kind
            of the domain of the process.
        key : hashable
            The name that identifies the new process in the GP object. If not
            specified, sets the default process.
        deriv : Deriv-like
            The linear combination is derived as specified by this
            parameter.
        
        """
        
        for k, func in ops.items():
            if k not in self._procs:
                raise KeyError(f'process key {k!r} not in GP object')
            if not jnp.isscalar(func) and not callable(func):
                raise TypeError(f'object of type {type(func)!r} for key {k!r} is neither scalar nor callable')
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = self._ProcTransf(ops, deriv)
        
        # we could implement addproctransf in terms of addproclintransf with
        # the following code, but addproctransf has linear kernel building
        # cost so I'm leaving it around (probably not significant anyway)
        # functions = [
        #     op if callable(op)
        #     else (lambda x: lambda _: x)(op)
        #     for op in ops.values()
        # ]
        # def equivalent_lintransf(*procs):
        #     def fun(x):
        #         out = None
        #         for fun, proc in zip(functions, procs):
        #             this = fun(x) * proc(x)
        #             out = this if out is None else out + this
        #         return out
        #     return fun
        # self.addproclintransf(equivalent_lintransf, list(ops.keys()), key, deriv, False)
    
    def addproclintransf(self, transf, keys, key=DefaultProcess, *, deriv=0, checklin=False):
        """
        
        Define a new process as a linear combination of other processes.
        
        Let f_i(x), i = 1, 2, ... be already defined processes, and T
        a linear map from processes to a single process. The new process is
        
            h(x) = T(f_1, f_2, ...)(x).
        
        Parameters
        ----------
        transf : callable
            A function with signature `transf(callable, callable, ...) -> callable`.
        keys : sequence
            The keys of the processes to be passed to the transformation.
        key : hashable
            The name that identifies the new process in the GP object. If not
            specified, sets the default process.
        deriv : Deriv-like
            The linear combination is derived as specified by this
            parameter.
        checklin : bool
            If True, check if the transformation is linear. Default False.
        
        Notes
        -----
        The linearity check may fail if the transformation does nontrivial
        operations with the inner function input.
        
        """
        
        # TODO support single key in keys
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
    
        for k in keys:
            if k not in self._procs:
                raise KeyError(k)
        
        deriv = _Deriv.Deriv(deriv)
        
        if len(keys) == 0:
            self._procs[key] = self._ProcKernel(self._zerokernel)
            return
        
        if checklin is None:
            checklin = self._checklin
        if checklin:
            mockup_function = lambda a: lambda _: a
            # TODO this array mockup fails with jax functions
            class Mockup(numpy.ndarray):
                __getitem__ = lambda *_: Mockup((0,))
                __getattr__ = __getitem__
            def checktransf(*arrays):
                functions = [mockup_function(a) for a in arrays]
                return transf(*functions)(Mockup((0,)))
            shapes = [(11,)] * len(keys)
            self._checklinear(checktransf, shapes, elementwise=True)
        
        self._procs[key] = self._ProcLinTransf(transf, keys, deriv)

    def addkernelop(self, method, arg, key, proc=DefaultProcess):
        """
        
        Define a new process as the transformation of an existing one.
        
        Parameters
        ----------
        method : str
            A method of `Kernel` taking two arguments which returns a
            transformed kernel.
        arg : object
            A valid argument to the method.
        key : hashable
            Key for the new process.
        proc : hashable
            Key of the process to be transformed. If not specified, use the
            default process.
        
        """
        
        if not hasattr(_Kernel.Kernel, method):
            raise ValueError(f'Kernel has not attribute {method!r}')
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        if proc not in self._procs:
            raise KeyError(f'process {proc!r} not found')
                
        self._procs[key] = self._ProcKernelOp(proc, method, arg)
    
    def addprocderiv(self, deriv, key, proc=DefaultProcess):
        """
        
        Define a new process as the derivative of an existing one.
        
        .. math::
            g(x) = \\frac{\\partial^n}{\\partial x^n} f(x)
        
        Parameters
        ----------
        deriv : Deriv-like
            Derivation order.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be derived. If not specified, use the
            default process.
        
        """
        deriv = _Deriv.Deriv(deriv)
        self.addkernelop('diff', deriv, key, proc)
    
    def addprocxtransf(self, transf, key, proc=DefaultProcess):
        """
        
        Define a new process by transforming the inputs of another one.
        
        .. math::
            g(x) = f(T(x))
        
        Parameters
        ----------
        transf : callable
            A function mapping the new kind input to the input expected by the
            transformed process.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be transformed. If not specified, use the
            default process.
        
        """
        assert callable(transf)
        self.addkernelop('xtransf', transf, key, proc)
    
    def addprocrescale(self, scalefun, key, proc=DefaultProcess):
        """
        
        Define a new process as a rescaling of an existing one.
        
        .. math::
            g(x) = s(x)f(x)
        
        Parameters
        ----------
        scalefun : callable
            A function from the domain of the process to a scalar.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be transformed. If not specified, use the
            default process.
        
        """
        assert callable(scalefun)
        self.addkernelop('rescale', scalefun, key, proc)
    
    def addx(self, x, key=None, *, deriv=0, proc=DefaultProcess):
        """
        
        Add points where the Gaussian process is evaluated.
        
        The GP object keeps the various x arrays in a dictionary. If `x` is an
        array, you have to specify its dictionary key with the `key` parameter.
        Otherwise, you can directly pass a dictionary for `x`.
        
        To specify that on the given `x` a derivative of the process instead of
        the process itself should be evaluated, use the parameter `deriv`.
        
        `addx` may or may not copy the input arrays.
        
        Parameters
        ----------
        x : array or dictionary of arrays
            The points to be added.
        key : hashable
            If `x` is an array, the dictionary key under which `x` is added.
            Can not be specified if `x` is a dictionary.
        deriv : Deriv-like
            Derivative specification. A :class:`Deriv` object or something that
            can be converted to :class:`Deriv`.
        proc : hashable
            The process to be evaluated on the points. If not specified, use
            the default process.
        
        """
        
        # TODO after I implement block solving, add per-key covariance matrix
        # flags.
        
        # TODO add `copy` parameter, default False, to copy the input arrays
        # if they are numpy arrays.
        
        # this interface does not allow adding a single dictionary as x element
        # unless it's wrapped as a 0d numpy array, but this is for the best
        
        deriv = _Deriv.Deriv(deriv)

        if proc not in self._procs:
            raise KeyError(f'process named {proc!r} not found')
                
        if hasattr(x, 'keys'):
            if key is not None:
                raise ValueError('can not specify key if x is a dictionary')
            if None in x:
                raise ValueError('None key in x not allowed')
        else:
            if key is None:
                raise ValueError('x is not dictionary but key is None')
            x = {key: x}
        
        for key in x:
            if key in self._elements:
                raise KeyError('key {!r} already in GP'.format(key))
            
            gx = x[key]
            
            # Convert to JAX array, numpy array or StructuredArray.
            # convert eagerly to jax to avoid problems with tracing.
            gx = _array._asarray_jaxifpossible(gx)

            # Check it is not empty.
            if not gx.size:
                raise ValueError('x[{!r}] is empty'.format(key))
                # TODO it would be elegant to support empty arrays, but how
                # many things would break down from here?

            # Check dtype is compatible with previous arrays.
            # TODO since we never concatenate arrays we could allow a less
            # strict compatibility. In principle we could allow really anything
            # as long as the kernel eats it, but this probably would let bugs
            # through without being really ever useful. What would make sense
            # is checking the dtype structure matches recursively and check
            # concrete dtypes of fields can be casted.
            # TODO result_type is too lax. Examples: str, float -> str,
            # object, float -> object. I should use something like the
            # ordering function in updowncast.py.
            if hasattr(self, '_dtype'):
                try:
                    self._dtype = numpy.result_type(self._dtype, gx.dtype)
                    # do not use jnp.result_type, it does not support
                    # structured types
                except TypeError:
                    msg = 'x[{!r}].dtype = {!r} not compatible with {!r}'
                    msg = msg.format(key, gx.dtype, self._dtype)
                    raise TypeError(msg)
            else:
                self._dtype = gx.dtype

            # Check that the derivative specifications are compatible with the
            # array data type.
            if gx.dtype.names is None:
                if not deriv.implicit:
                    raise ValueError('x has no fields but derivative has')
            else:
                for dim in deriv:
                    if dim not in gx.dtype.names:
                        raise ValueError(f'deriv field {dim!r} not in x')
            
            self._elements[key] = self._Points(gx, deriv, proc)

    def _get_x_dtype(self):
        """ Get the data type of x points """
        return getattr(self, '_dtype', None)
        
    def addtransf(self, tensors, key, *, axes=1):
        """
        
        Apply a linear transformation to already specified process points. The
        result of the transformation is represented by a new key.
        
        Parameters
        ----------
        tensors : dict
            Dictionary mapping keys of the GP to arrays/scalars. Each array is
            matrix-multiplied with the process array represented by its key,
            while scalars are just multiplied. Finally, the keys are summed
            over.
        key : hashable
            A new key under which the transformation is placed.
        axes : int
            Number of axes to be summed over for matrix multiplication,
            referring to trailing axes for tensors in `tensors`, and to
            heading axes for process points. Default 1.
        
        Notes
        -----
        The multiplication between the tensors and the process is done with
        np.tensordot with, by default, 1-axis contraction. For >2d arrays this
        is different from numpy's matrix multiplication, which would act on the
        second-to-last dimension of the second array.
        
        """        
        # Note: it may seem nice that when an array has less axes than `axes`,
        # the summation would be restricted only on the existing axes. However
        # this brings about the ambiguous case where only one of the factors has
        # not enough axes. How many axes do you sum over on the other?
        
        # Check axes.
        assert isinstance(axes, int) and axes >= 0, axes
        
        # Check key.
        if key is None:
            raise ValueError('key can not be None')
        if key in self._elements:
            raise KeyError(f'key {key!r} already in GP')
        
        # Check keys.
        for k in tensors:
            if k not in self._elements:
                raise KeyError(k)
        
        # Check tensors and convert them to jax arrays.
        if len(tensors) == 0:
            raise ValueError('empty tensors, undetermined output shape')
        tens = {}
        for k, t in tensors.items():
            t = jnp.asarray(t)
            # no need to check dtype since jax supports only numerical arrays
            with _patch_jax.skipifabstract():
                if self._checkfinite and not jnp.all(jnp.isfinite(t)):
                    raise ValueError(f'tensors[{k!r}] contains infs/nans')
            rshape = self._elements[k].shape
            if t.shape and t.shape[t.ndim - axes:] != rshape[:axes]:
                raise ValueError(f'tensors[{k!r}].shape = {t.shape!r} can not be multiplied with shape {rshape!r} with {axes}-axes contraction')
            tens[k] = t
        
        # Check shapes broadcast correctly.
        arrays = tens.values()
        elements = (self._elements[k] for k in tens)
        shapes = (
            t.shape[:t.ndim - axes] + e.shape[axes:] if t.shape else e.shape
            for t, e in zip(arrays, elements)
        )
        try:
            shape = jnp.broadcast_shapes(*shapes)
        except ValueError:
            msg = 'can not broadcast tensors with shapes ['
            msg += ', '.join(repr(t.shape) for t in arrays)
            msg += '] contracted with arrays with shapes ['
            msg += ', '.join(repr(e.shape) for e in elements) + ']'
            raise ValueError(msg)
        
        # Define linear transformation.
        def equiv_lintransf(*args):
            assert len(args) == len(tens)
            out = None
            for a, (k, t) in zip(args, tens.items()):
                if t.shape:
                    b = jnp.tensordot(t, a, axes)
                else:
                    b = t * a
                if out is None:
                    out = b
                else:
                    out = out + b
            return out
        keys = list(tens.keys())
        self.addlintransf(equiv_lintransf, keys, key, checklin=False)
    
    def addlintransf(self, transf, keys, key, *, checklin=None):
        """
        
        Define a finite linear transformation of the evaluated process.
        
        Parameters
        ----------
        transf : callable
            A function with signature `f(array1, array2, ...) -> array` which
            computes the linear transformation. The function must be
            jax-traceable, i.e., use jax.numpy instead of numpy.
        keys : sequence
            Keys of parts of the process to be passed as inputs to the
            transformation.
        key : hashable
            The key of the newly defined points.
        checklin : bool
            If True (default), check that the given function is linear in its
            inputs. The default can be overridden at initialization of the GP
            object. Note that an affine function (x -> a + bx) is not linear.
        
        Raises
        ------
        RuntimeError :
            The transformation seems not to be linear. To disable the linearity
            check, initialize the GP with `checklin=False`.
        
        """
        
        # TODO elementwise operations can be applied more efficiently to
        # primary gvars (tipical case), so the method could use an option
        # `elementwise`. What is the reliable way to check it is indeed
        # elementwise with a single random vector? Zero items of the tangent
        # at random with p=0.5 and check they stay zero? (And of course check
        # the shape is preserved.)
        
        # Check key.
        if key is None:
            raise ValueError('key can not be None')
        if key in self._elements:
            raise KeyError(f'key {key!r} already in GP')
        
        # Check keys.
        for k in keys:
            if k not in self._elements:
                raise KeyError(k)
        
        # Determine shape.
        class ArrayMockup:
            def __init__(self, elem):
                self.shape = elem.shape
                self.dtype = float
        inp = [ArrayMockup(self._elements[k]) for k in keys]
        out = jax.eval_shape(transf, *inp)
        shape = out.shape
        
        # Check that the transformation is linear.
        if checklin is None:
            checklin = self._checklin
        if checklin:
            shapes = [self._elements[k].shape for k in keys]
            self._checklinear(transf, shapes)
        
        self._elements[key] = self._LinTransf(transf, keys, shape)
    
    def _checklinear(self, func, inshapes, elementwise=False):
        
        # Make input arrays.
        rkey = jax.random.PRNGKey(202206091600)
        inp = []
        for shape in inshapes:
            rkey, subkey = jax.random.split(rkey)
            inp.append(jax.random.normal(subkey, shape))
        
        # Put zeros into the arrays to check they are preserved.
        if elementwise:
            shape = jnp.broadcast_shapes(*inshapes)
            rkey, subkey = jax.random.split(rkey)
            zeros = jax.random.bernoulli(subkey, 0.5, shape)
            for i, a in enumerate(inp):
                inp[i] = a.at[zeros].set(0)
            
        # Compute JVP and check it is identical to the function itself.
        with _patch_jax.skipifabstract():
            out0, out1 = jax.jvp(func, inp, inp)
            if out1.dtype == jax.float0:
                cond = jnp.allclose(out0, 0)
            else:
                cond = jnp.allclose(out0, out1)
            if not cond:
                raise RuntimeError('the transformation is not linear')
        
            # Check that the function is elementwise.
            if elementwise:
                if out0.shape != shape or not (jnp.allclose(out0[zeros], 0) and jnp.allclose(out1[zeros], 0)):
                    raise RuntimeError('the transformation is not elementwise')
    
    def addcov(self, covblocks, key=None, *, decomps=None):
        """
        
        Add user-defined prior covariance matrix blocks.
        
        Covariance matrices defined with `addcov` represent arbitrary
        finite-dimensional zero-mean Gaussian variables, assumed independent
        from all other variables in the GP object.
                
        Parameters
        ----------
        covblocks : array or dictionary of arrays
            If an array: a covariance matrix (or tensor) to be added under key
            `key`. If a dictionary: a mapping from pairs of keys to the
            corresponding covariance matrix blocks. A missing off-diagonal
            block in the dictionary is interpreted as a matrix of zeros,
            unless the corresponding transposed block is specified.
        key : hashable
            If `covblocks` is an array, the dictionary key under which
            `covblocks` is added. Can not be specified if `covblocks` is a
            dictionary.
        decomps : Decomposition or dict of Decompositions
            Pre-computed decompositions of (not necessarily all) diagonal
            blocks, as produced by :meth:`~GP.decompose`. The keys are single
            GP keys and not pairs like in `covblocks`.
        
        Raises
        ------
        KeyError :
            A key is already used in the GP.
        ValueError :
            `covblocks` and/or `key` and `decomps` are malformed or
            inconsistent.
        TypeError :
            Wrong type of `covblocks` or `decomps`.
        
        """
        
        # TODO maybe allow passing only the lower/upper triangular part for
        # the diagonal blocks, like I meta-allow for out of diagonal blocks?
        
        # TODO with multiple blocks and a single decomp, the decomp could be
        # interpreted as the decomposition of the whole block matrix.
        
        # Check type of `covblocks` and standardize it to dictionary.
        if hasattr(covblocks, 'keys'):
            if key is not None:
                raise ValueError('can not specify key if covblocks is a dictionary')
            if None in covblocks:
                raise ValueError('None key in covblocks not allowed')
            if decomps is not None and not hasattr(decomps, 'keys'):
                raise TypeError('covblocks is dictionary but decomps is not')
        else:
            if key is None:
                raise ValueError('covblocks is not dictionary but key is None')
            covblocks = {(key, key): covblocks}
            if decomps is not None:
                decomps = {key: decomps}
        
        if decomps is None:
            decomps = {}
        
        # Convert blocks to jax arrays and determine shapes from diagonal
        # blocks.
        shapes = {}
        preblocks = {}
        for keys, block in covblocks.items():
            # TODO maybe check that keys is a 2-tuple
            for key in keys:
                if key in self._elements:
                    raise KeyError(f'key {key!r} already in GP')
            xkey, ykey = keys
            if block is None:
                raise TypeError(f'block {keys!r} is None')
                # because jnp.asarray(None) interprets None as nan
                # (see jax issue #14506)
            block = jnp.asarray(block)
            
            if xkey == ykey:
                
                if block.ndim % 2 == 1:
                    raise ValueError(f'diagonal block {key!r} has odd number of axes')
                
                half = block.ndim // 2
                head = block.shape[:half]
                tail = block.shape[half:]
                if head != tail:
                    raise ValueError(f'shape {block.shape!r} of diagonal block {key!r} is not symmetric')
                shapes[xkey] = head
                
                with _patch_jax.skipifabstract():
                    if self._checksym and not jnp.allclose(block, block.T):
                        raise ValueError(f'diagonal block {key!r} is not symmetric')
                
            preblocks[keys] = block
        
        # Check decomps is consistent with covblocks.
        for key, dec in decomps.items():
            if key not in shapes:
                raise KeyError(f'key {key!r} in decomps not found in diagonal blocks')
            if not isinstance(dec, _linalg.Decomposition):
                raise TypeError(f'decomps[{key!r}] = {dec!r} is not a decomposition')
            n = numpy.prod(shapes[key], dtype=int)
            if dec.n != n:
                raise ValueError(f'decomposition matrix size {dec.n} != diagonal block size {n} for key {key!r}')
        
        # Reshape blocks to square matrices and check that the shapes of out of
        # diagonal blocks match those of diagonal ones.
        blocks = {}
        for keys, block in preblocks.items():
            with _patch_jax.skipifabstract():
                if self._checkfinite and not jnp.all(jnp.isfinite(block)):
                    raise ValueError(f'block {keys!r} not finite')
            xkey, ykey = keys
            if xkey == ykey:
                size = numpy.prod(shapes[xkey], dtype=int)
                blocks[keys] = block.reshape((size, size))
            else:
                for key in keys:
                    if key not in shapes:
                        raise KeyError(f'key {key!r} from off-diagonal block {keys!r} not found in diagonal blocks')
                eshape = shapes[xkey] + shapes[ykey]
                if block.shape != eshape:
                    raise ValueError(f'shape {block.shape!r} of block {keys!r} is not {eshape!r} as expected from diagonal blocks')
                xsize = numpy.prod(shapes[xkey], dtype=int)
                ysize = numpy.prod(shapes[ykey], dtype=int)
                block = block.reshape((xsize, ysize))
                blocks[keys] = block
                revkeys = keys[::-1]
                blockT = preblocks.get(revkeys)
                if blockT is None:
                    blocks[revkeys] = block.T
        
        # Check symmetry of out of diagonal blocks.
        if self._checksym:
            with _patch_jax.skipifabstract():
                for keys, block in blocks.items():
                    xkey, ykey = keys
                    if xkey != ykey:
                        blockT = blocks[ykey, xkey]
                        if not jnp.allclose(block.T, blockT):
                            raise ValueError(f'block {keys!r} is not the transpose of block {revkeys!r}')
        
        # Create _Cov objects.
        for key, shape in shapes.items():
            self._elements[key] = self._Cov(blocks, shape)
            decomp = decomps.get(key)
            if decomp is not None:
                self._decompcache[key,] = decomp

    def _crosskernel(self, xpkey, ypkey):
        
        # Check if the kernel is in cache.
        cache = self._kernels.get((xpkey, ypkey))
        if cache is not None:
            return cache
        
        # Compute the kernel.
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if isinstance(xp, self._ProcKernel) and isinstance(yp, self._ProcKernel):
            kernel = self._crosskernel_kernels(xpkey, ypkey)
        elif isinstance(xp, self._ProcTransf):
            kernel = self._crosskernel_transf_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcTransf):
            kernel = self._crosskernel_transf_any(ypkey, xpkey)._swap()
        elif isinstance(xp, self._ProcLinTransf):
            kernel = self._crosskernel_lintransf_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcLinTransf):
            kernel = self._crosskernel_lintransf_any(ypkey, xpkey)._swap()
        elif isinstance(xp, self._ProcKernelOp):
            kernel = self._crosskernel_op_any(xpkey, ypkey)
        elif isinstance(yp, self._ProcKernelOp):
            kernel = self._crosskernel_op_any(ypkey, xpkey)._swap()
        else:
            raise TypeError(f'unrecognized process types {type(xp)!r} and {type(yp)!r}')
        
        # Save cache.
        self._kernels[xpkey, ypkey] = kernel
        self._kernels[ypkey, xpkey] = kernel._swap()
        
        return kernel
        
    def _crosskernel_kernels(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            return xp.kernel.diff(xp.deriv, xp.deriv)
        else:
            return self._zerokernel
    
    def _crosskernel_transf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        kernelsum = self._zerokernel
        
        for pkey, factor in xp.ops.items():
            kernel = self._crosskernel(pkey, ypkey)
            if kernel is self._zerokernel:
                continue
            
            if not callable(factor):
                factor = (lambda f: lambda _: f)(factor)
            kernel = kernel.rescale(factor, None)
            
            if kernelsum is self._zerokernel:
                kernelsum = kernel
            else:
                kernelsum += kernel
                
        return kernelsum.diff(xp.deriv, 0)
    
    def _crosskernel_lintransf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        kernels = [self._crosskernel(pk, ypkey) for pk in xp.keys]
        kernel = _Kernel.CrossKernel._nary(xp.transf, kernels, _Kernel.CrossKernel._side.LEFT)
        kernel = kernel.diff(xp.deriv, 0)
        
        return kernel
    
    def _crosskernel_op_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            basekernel = self._crosskernel(xp.proc, xp.proc)
            # In principle I could avoid handling this case separately but it
            # will probably allow simplifications in how I implement nontrivial
            # transformations in Kernel: I won't need to support taking a
            # transformation in two steps.
        else:
            basekernel = self._crosskernel(xp.proc, ypkey)
        
        if basekernel is self._zerokernel:
            return self._zerokernel
        elif xp is yp:
            return getattr(basekernel, xp.method)(xp.arg, xp.arg)
        else:
            return getattr(basekernel, xp.method)(xp.arg, None)
    
    def _makecovblock_points(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        
        assert isinstance(x, self._Points)
        assert isinstance(y, self._Points)
        
        kernel = self._crosskernel(x.proc, y.proc)
        if kernel is self._zerokernel:
            # TODO handle zero cov block efficiently
            return jnp.zeros((x.size, y.size))
        
        kernel = kernel.diff(x.deriv, y.deriv)
        
        if x is y and not self._checksym and self._halfmatrix:
            ix, iy, back = self._triu_indices_and_back(x.size)
            flat = x.x.reshape(-1)
            ax = flat[ix]
            ay = flat[iy]
            halfcov = kernel(ax, ay)
            cov = halfcov[back]
            # TODO to avoid inefficiencies like in BART, maybe _Kernel should
            # have a method outer(x) that by default simply does self(x[None,
            # :], x[:, None]) but can be overwritten. This halfmatrix impl could
            # be moved there with an option outer(x, *, half=False). To carry
            # over custom implementations of outer, there should be a callable
            # attribute _outer, optionally set at initialization, that is
            # transformed by kernel operations.
        else:
            ax = x.x.reshape(-1)[:, None]
            ay = y.x.reshape(-1)[None, :]
            cov = kernel(ax, ay)
        
        return cov
    
    def _makecovblock_lintransf_any(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        assert isinstance(x, self._LinTransf)
        
        # Gather covariance matrices to be transformed.
        covs = []
        for k in x.keys:
            elem = self._elements[k]
            cov = self._covblock(k, ykey)
            assert cov.shape == (elem.size, y.size)
            cov = cov.reshape(elem.shape + (y.size,))
            covs.append(cov)
        
        # Apply transformation.
        t = jax.vmap(x.transf, -1, -1)
        cov = t(*covs)
        assert cov.shape == x.shape + (y.size,)
        return cov.reshape((x.size, y.size)) # don't leave out the ()!
        # the () probably was an obscure autograd bug, I don't think it will
        # be a problem again with jax
            
    def _makecovblock(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        if isinstance(x, self._Points) and isinstance(y, self._Points):
            cov = self._makecovblock_points(xkey, ykey)
        elif isinstance(x, self._LinTransf):
            cov = self._makecovblock_lintransf_any(xkey, ykey)
        elif isinstance(y, self._LinTransf):
            cov = self._makecovblock_lintransf_any(ykey, xkey)
            cov = cov.T
        elif isinstance(x, self._Cov) and isinstance(y, self._Cov) and x.blocks is y.blocks and (xkey, ykey) in x.blocks:
            cov = x.blocks[xkey, ykey]
        else:
            # TODO handle zero cov block efficiently
            cov = jnp.zeros((x.size, y.size))
        
        with _patch_jax.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(cov)):
                raise RuntimeError(f'covariance block {(xkey, ykey)!r} is not finite')
            if self._checksym and xkey == ykey and not jnp.allclose(cov, cov.T):
                raise RuntimeError(f'covariance block {(xkey, ykey)!r} is not symmetric')

        return cov
    
    def _covblock(self, row, col):
        
        if (row, col) not in self._covblocks:
            block = self._makecovblock(row, col)
            if row != col:
                if self._checksym:
                    with _patch_jax.skipifabstract():
                        blockT = self._makecovblock(col, row)
                        if not jnp.allclose(block.T, blockT):
                            msg = 'covariance block {!r} is not symmetric'
                            raise RuntimeError(msg.format((row, col)))
                self._covblocks[col, row] = block.T
            self._covblocks[row, col] = block

        return self._covblocks[row, col]
        
    def _assemblecovblocks(self, rowkeys, colkeys=None):
        if colkeys is None:
            colkeys = rowkeys
        blocks = [
            [self._covblock(row, col) for col in colkeys]
            for row in rowkeys
        ]
        return jnp.block(blocks)
    
    def _solver(self, keys, ycov=None, *, covtransf=None, **kw):
        """
        Return a decomposition of the covariance matrix of the keys in `keys`
        plus the matrix ycov. Keyword arguments are passed to the decomposition.
        """
        
        # TODO cache ignores **kw.

        keys = tuple(keys)
        
        # Check if decomposition is in cache.
        if ycov is None:
            cache = self._decompcache.get(keys)
            if cache is not None:
                return cache
            # TODO use frozenset(keys) instead of tuple(keys) to make cache
            # work when order changes, but I have to permute the decomposition
            # to make that work. Needs an ad-hoc class in _linalg. Make the
            # decompcache a dict subclass that accepts tuples of keys but uses
            # internally frozenset.
        
        # Compute decomposition.
        if isinstance(ycov, _linalg.Decomposition):
            ancestors = []
            transfs = []
            for key in keys:
                elem = self._elements[key]
                nest = False
                if isinstance(elem, self._LinTransf):
                    size = sum(self._elements[k].size for k in elem.keys)
                    if size < elem.size:
                        nest = True
                        ancestors += list(elem.keys)
                        transfs.append(jnp.concatenate(elem.matrices(self), 1))
                if not nest:
                    ancestors.append(key)
                    transfs.append(jnp.eye(elem.size))
            transf = jlinalg.block_diag(*transfs)
            cov = self._assemblecovblocks(ancestors)
            if covtransf:
                ycov, transf, cov = covtransf((ycov, transf, cov))
            covdec = self._decompclass(cov, **kw)
            # TODO obtain covdec from _solver recursively, to use cache?
            decomp = _linalg.Woodbury2(ycov, transf, covdec, self._decompclass, sign=1, **kw)
        else:
            Kxx = self._assemblecovblocks(keys)
            if ycov is not None:
                Kxx = Kxx + ycov
            if covtransf:
                Kxx = covtransf(Kxx)
            decomp = self._decompclass(Kxx, **kw)
        
        # Cache decomposition.
        if ycov is None:
            self._decompcache[keys] = decomp
        
        return decomp
        
    def _checkpos(self, cov):
        with _patch_jax.skipifabstract():
            # eigv = jnp.linalg.eigvalsh(cov)
            # mineigv, maxeigv = jnp.min(eigv), jnp.max(eigv)
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', r'Exited at iteration .+? with accuracies')
                warnings.filterwarnings('ignore', r'Exited postprocessing with accuracies')
                X = numpy.random.randn(len(cov), 1)
                A = numpy.asarray(cov)
                (mineigv,), _ = sparse.linalg.lobpcg(A, X, largest=False)
                (maxeigv,), _ = sparse.linalg.lobpcg(A, X, largest=True)
            assert mineigv <= maxeigv
            if mineigv < 0:
                bound = -len(cov) * jnp.finfo(cov.dtype).eps * maxeigv * self._posepsfac
                if mineigv < bound:
                    msg = 'covariance matrix is not positive definite: '
                    msg += 'mineigv = {:.4g} < {:.4g}'.format(mineigv, bound)
                    raise numpy.linalg.LinAlgError(msg)
    
    _checkpos_cache = functools.cached_property(lambda self: [])
    def _checkpos_keys(self, keys):
        # TODO go back to ancestors of _LinTransf?
        if not self._checkpositive:
            return
        keys = set(keys)
        for prev_keys in self._checkpos_cache:
            if keys.issubset(prev_keys):
                return
        cov = self._assemblecovblocks(list(keys))
        self._checkpos(cov)
        self._checkpos_cache.append(keys)
    
    def _priorpointscov(self, key):
        
        x = self._elements[key]
        classes = (self._Points, self._Cov)
        assert isinstance(x, classes)
        mean = numpy.zeros(x.size)
        cov = self._covblock(key, key).astype(float)
        assert cov.shape == 2 * mean.shape, cov.shape
        
        # get preexisting primary gvars to be correlated with the new ones
        preitems = [
            k
            for k, px in self._elements.items()
            if isinstance(px, classes)
            and k in self._priordict
        ]
        if preitems:
            prex = numpy.concatenate([
                numpy.reshape(self._priordict[k], -1)
                for k in preitems
            ])
            precov = numpy.concatenate([
                self._covblock(k, key).astype(float)
                for k in preitems
            ])
            g = gvar.gvar(mean, cov, prex, precov, fast=True)
        else:
            g = gvar.gvar(mean, cov, fast=True)
        
        return g.reshape(x.shape)
    
    def _priorlintransf(self, key):
        x = self._elements[key]
        assert isinstance(x, self._LinTransf)
        
        # Gather all gvars to be transformed.
        elems = [
            self._prior(k).reshape(-1)
            for k in x.keys
        ]
        g = numpy.concatenate(elems)
        
        # Extract jacobian and split it.
        slices = self._slices(x.keys)
        jac, indices = _patch_gvar.jacobian(g)
        jacs = [
            jac[s].reshape(self._elements[k].shape + indices.shape)
            for s, k in zip(slices, x.keys)
        ]
        # TODO the jacobian can be extracted much more efficiently when the
        # elements are _Points or _Cov, since in that case the gvars are primary
        # and contiguous within each block, so each jacobian is the identity + a
        # range. Then write a function _patch_gvar.merge_jacobians to combine
        # them, which also can be optimized knowing the indices are
        # non-overlapping ranges.
        
        # Apply transformation.
        t = jax.vmap(x.transf, -1, -1)
        outjac = t(*jacs)
        assert outjac.shape == x.shape + indices.shape
        
        # Rebuild gvars.
        outg = _patch_gvar.from_jacobian(numpy.zeros(x.shape), outjac, indices)
        return outg
    
    def _prior(self, key):
        prior = self._priordict.get(key, None)
        if prior is None:
            x = self._elements[key]
            if isinstance(x, (self._Points, self._Cov)):
                prior = self._priorpointscov(key)
            elif isinstance(x, self._LinTransf):
                prior = self._priorlintransf(key)
            else: # pragma: no cover
                raise TypeError(type(x))
            self._priordict[key] = prior
        return prior
    
    def prior(self, key=None, *, raw=False):
        """
        
        Return an array or a dictionary of arrays of gvars representing the
        prior for the Gaussian process. The returned object is not unique but
        the gvars stored inside are, so all the correlations are kept between
        objects returned by different calls to `prior`.
        
        Calling without arguments returns the complete prior as a dictionary.
        If you specify `key`, only the array for the requested key is returned.
        
        Parameters
        ----------
        key : None, key or list of keys
            Key(s) corresponding to one passed to `addx` or `addtransf`. None
            for all keys.
        raw : bool
            If True, instead of returning a collection of gvars return
            their covariance matrix as would be returned by `gvar.evalcov`.
            Default False.
        
        Returns
        -------
        If raw=False (default):
        
        prior : np.ndarray or gvar.BufferDict
            A collection of gvars representing the prior.
        
        If raw=True:
        
        cov : np.ndarray or dict
            The covariance matrix of the prior.
        """
        raw = bool(raw)
        
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = None
        
        self._checkpos_keys([key] if outkeys is None else outkeys)
        
        if raw and outkeys is not None:
            return {
                (row, col):
                self._covblock(row, col).reshape(
                    self._elements[row].shape +
                    self._elements[col].shape
                )
                for row in outkeys
                for col in outkeys
            }
        elif raw:
            return self._covblock(key, key).reshape(2 * self._elements[key].shape)
        elif outkeys is not None:
            return gvar.BufferDict({
                key: self._prior(key) for key in outkeys
            })
        else:
            return self._prior(key)
        
    def _flatgiven(self, given, givencov):
        
        if not hasattr(given, 'keys'):
            raise TypeError('`given` must be dict')
        gcblack = givencov is None or isinstance(givencov, _linalg.Decomposition)
        if not gcblack and not hasattr(givencov, 'keys'):
            raise TypeError('`givenconv` must be None, dict or Decomposition')
        
        ylist = []
        keylist = []
        for key, l in given.items():
            if key not in self._elements:
                raise KeyError(key)
            
            if not isinstance(l, jnp.ndarray):
                # use numpy since there could be gvars
                l = numpy.asarray(l)
            shape = self._elements[key].shape
            if l.shape != shape:
                msg = 'given[{!r}] has shape {!r} different from shape {!r}'
                raise ValueError(msg.format(key, l.shape, shape))
            if l.dtype != object and not jnp.issubdtype(l.dtype, jnp.number):
                msg = 'given[{!r}] has non-numerical dtype {!r}'
                raise TypeError(msg.format(key, l.dtype))
            
            ylist.append(l.reshape(-1))
            keylist.append(key)
        
        # TODO error checking on the unpacking of givencov
        
        if gcblack:
            covblocks = givencov
        else:
            covblocks = [
                [
                    jnp.asarray(givencov[keylist[i], keylist[j]]).reshape(ylist[i].shape + ylist[j].shape)
                    for j in range(len(keylist))
                ]
                for i in range(len(keylist))
            ]
            
        return ylist, keylist, covblocks
    
    def _slices(self, keylist):
        """
        Return list of slices for the positions of flattened arrays
        corresponding to keys in `keylist` into their concatenation.
        """
        sizes = [self._elements[key].size for key in keylist]
        stops = numpy.pad(numpy.cumsum(sizes), (1, 0))
        return [slice(stops[i - 1], stops[i]) for i in range(1, len(stops))]
    
    def pred(self, given, key=None, givencov=None, *, fromdata=None, raw=False, keepcorr=None):
        """
        
        Compute the posterior.
        
        The posterior can be computed either for all points or for a subset,
        and either directly from data or from a posterior obtained with a fit.
        The latter case is for when the Gaussian process was used in a fit with
        other parameters.
        
        The output is a collection of gvars, either an array or a dictionary
        of arrays. They are properly correlated with gvars returned by
        `prior` and with the input data/fit.
        
        The input is a dictionary of arrays, `given`, with keys corresponding
        to the keys in the GP as added by `addx` or `addtransf`.
        
        Parameters
        ----------
        given : dictionary of arrays
            The data or fit result for some/all of the points in the GP.
            The arrays can contain either gvars or normal numbers, the latter
            being equivalent to zero-uncertainty gvars.
        key : None, key or list of keys, optional
            If None, compute the posterior for all points in the GP (also those
            used in `given`). Otherwise only those specified by key.
        givencov : dictionary of arrays, optional
            Covariance matrix of `given`. If not specified, the covariance
            is extracted from `given` with `gvar.evalcov(given)`.
        fromdata : bool
            Mandatory. Specify if the contents of `given` are data or already
            a posterior.
        raw : bool, optional
            If True, instead of returning a collection of gvars, return
            the mean and the covariance. When the mean is a dictionary, the
            covariance is a dictionary whose keys are pairs of keys of the
            mean (the same format used by `gvar.evalcov`). Default False.
        keepcorr : bool, optional
            If True (default), the returned gvars are correlated with the
            prior and the data/fit. If False, they have the correct covariance
            between themselves, but are independent from all other preexisting
            gvars.
        
        Returns
        -------
        If raw=False (default):
        
        posterior : array or dictionary of arrays
            A collections of gvars representing the posterior.
        
        If raw=True:
        
        pmean : array or dictionary of arrays
            The mean of the posterior. Equivalent to `gvar.mean(posterior)`.
        pcov : 2D array or dictionary of 2D arrays
            The covariance matrix of the posterior. If `pmean` is a dictionary,
            the keys of `pcov` are pairs of keys of `pmean`. Equivalent to
            `gvar.evalcov(posterior)`.
        
        """

        # TODO GP.pred(..., raw=True, onlyvariance=True) computes only the
        # variance (requires actually implementing diagquad at least in Chol and
        # Diag).

        
        if fromdata is None:
            raise ValueError('you must specify if `given` is data or fit result')
        fromdata = bool(fromdata)
        raw = bool(raw)
        if keepcorr is None:
            keepcorr = not raw
        if keepcorr and raw:
            raise ValueError('both keepcorr=True and raw=True')
        
        strip = False
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = [key]
            strip = True
        outslices = self._slices(outkeys)
                
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = self._concatenate(ylist)
        if y.dtype == object:
            if ycovblocks is not None:
                raise ValueError('given may contain gvars but a separate covariance matrix has been provided')
        
        self._checkpos_keys(inkeys + outkeys)
        
        Kxxs = self._assemblecovblocks(inkeys, outkeys)
        
        if isinstance(ycovblocks, _linalg.Decomposition):
            ycov = ycovblocks
        elif ycovblocks is not None:
            ycov = jnp.block(ycovblocks)
        elif (fromdata or raw or not keepcorr) and y.dtype == object:
            ycov = gvar.evalcov(gvar.gvar(y))
            # TODO use evalcov_blocks
        else:
            ycov = None
        self._check_ycov(ycov)
        
        if raw or not keepcorr or self._checkfinite:
            if y.dtype == object:
                ymean = gvar.mean(y)
            else:
                ymean = y
            self._check_ymean(ymean)
        
        if raw or not keepcorr:
            
            Kxsxs = self._assemblecovblocks(outkeys)
            
            if fromdata:
                solver = self._solver(inkeys, ycov)
            else:
                solver = self._solver(inkeys)

            mean = solver.quad(Kxxs, ymean)
            cov = Kxsxs - solver.quad(Kxxs)
            
            if not fromdata:
                # complete formula:
                # A = Kxx^-1 @ Kxxs
                # cov = Kxsxs - A.T @ (Kxx - ycov) @ A
                if ycov is not None:
                    if isinstance(ycov, _linalg.Decomposition):
                        ycov = ycov.matrix()
                    A = solver.solve(Kxxs)
                    cov = cov + A.T @ ycov @ A
            
        else: # (keepcorr and not raw)        
            yplist = [numpy.reshape(self._prior(key), -1) for key in inkeys]
            ysplist = [numpy.reshape(self._prior(key), -1) for key in outkeys]
            yp = self._concatenate(yplist)
            ysp = self._concatenate(ysplist)
            
            if y.dtype != object and ycov is not None:
                if isinstance(ycov, _linalg.Decomposition):
                    ycov = ycov.matrix()
                y = gvar.gvar(y, ycov)
            else:
                y = numpy.asarray(y) # because y - yp fails if y is a jax array
            mat = ycov if fromdata else None
            flatout = ysp + self._solver(inkeys, mat).quad(Kxxs, y - yp)
        
        if raw and not strip:
            meandict = {
                key: mean[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            }
            
            covdict = {
                (row, col):
                cov[rowslice, colslice].reshape(self._elements[row].shape + self._elements[col].shape)
                for row, rowslice in zip(outkeys, outslices)
                for col, colslice in zip(outkeys, outslices)
            }
            
            return meandict, covdict
            
        elif raw:
            outkey, = outkeys
            mean = mean.reshape(self._elements[outkey].shape)
            cov = cov.reshape(2 * self._elements[outkey].shape)
            return mean, cov
        
        elif not keepcorr:
                        
            flatout = gvar.gvar(mean, cov, fast=True)
        
        if not strip:
            return gvar.BufferDict({
                key: flatout[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            })
        else:
            outkey, = outkeys
            return flatout.reshape(self._elements[outkey].shape)
        
    def predfromfit(self, *args, **kw):
        """
        Like `pred` with `fromdata=False`.
        """
        return self.pred(*args, fromdata=False, **kw)
    
    def predfromdata(self, *args, **kw):
        """
        Like `pred` with `fromdata=True`.
        """
        return self.pred(*args, fromdata=True, **kw)
    
    def _prior_decomp(self, given, givencov=None, **kw):
        """ Internal implementation of marginal_likelihood. Keyword arguments
        are passed to _solver. """
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = self._concatenate(ylist)
        
        self._checkpos_keys(inkeys)
        
        # Get mean.
        if y.dtype == object:
            ymean = gvar.mean(y)
        else:
            ymean = y
        self._check_ymean(ymean)
        
        # Get covariance matrix.
        if isinstance(ycovblocks, _linalg.Decomposition):
            ycov = ycovblocks
        elif ycovblocks is not None:
            ycov = jnp.block(ycovblocks)
            if y.dtype == object:
                warnings.warn(f'covariance matrix may have been specified both explicitly and with gvars; the explicit one will be used')
        elif y.dtype == object:
            gvary = gvar.gvar(y)
            ycov = gvar.evalcov(gvary)
        else:
            ycov = None
        self._check_ycov(ycov)
        
        decomp = self._solver(inkeys, ycov, **kw)
        return decomp, ymean
    
    def _check_ymean(self, ymean):
        with _patch_jax.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(ymean)):
                raise ValueError('mean of `given` is not finite')
    
    def _check_ycov(self, ycov):
        if ycov is None or isinstance(ycov, _linalg.Decomposition):
            return
        with _patch_jax.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(ycov)):
                raise ValueError('covariance matrix of `given` is not finite')
            if self._checksym and not jnp.allclose(ycov, ycov.T):
                raise ValueError('covariance matrix of `given` is not symmetric')

    def marginal_likelihood(self, given, givencov=None, *, separate=False, **kw):
        """
        
        Compute the logarithm of the probability of the data.
        
        The probability is computed under the Gaussian prior and Gaussian error
        model. It is also called marginal likelihood. If `y` is the data and
        `g` is the Gaussian process, this is
        
        .. math::
            \\log \\int p(y|g) p(g) \\mathrm{d} g.
        
        Unlike `pred()`, you can't compute this with a fit result instead of
        data. If you used the Gaussian process as latent variable in a fit,
        use the whole fit to compute the marginal likelihood. E.g. `lsqfit`
        always computes the logGBF (it's the same thing).
        
        The input is an array or dictionary of arrays, `given`. The contents of
        `given` represent the input data.
                
        Parameters
        ----------
        given : dictionary of arrays
            The data for some/all of the points in the GP. The arrays can
            contain either gvars or normal numbers, the latter being
            equivalent to zero-uncertainty gvars.
        givencov : dictionary of arrays, optional
            Covariance matrix of `given`. If not specified, the covariance
            is extracted from `given` with `gvar.evalcov(given)`.
        separate : bool
            If True, return separately the logdet term and the residuals term.
            Default False.
        **kw :
            Additional keyword arguments are passed to the matrix decomposition.
        
        Returns
        -------
        If ``separate == False`` (default):
        
        logp : scalar
            The logarithm of the marginal likelihood.
        
        If ``separate == True``:
            
        logdet : scalar
            The term of the marginal likelihood containing the logarithm of
            the determinant of the prior covariance matrix, without the -1/2
            factor.
        residuals : array or dictionary of arrays
            A vector whose squared 2-norm multiplied by -1/2 gives the other
            term of the marginal likelihood.
        """
        decomp, ymean = self._prior_decomp(given, givencov, **kw)
        logdet = decomp.logdet() + decomp.n * jnp.log(2 * jnp.pi)
        if separate:
            residuals = decomp.decorrelate(ymean)
            return logdet, residuals
        else:
            return -1/2 * (decomp.quad(ymean) + logdet)
    
    @staticmethod
    def _getdecomp(solver):
        return {
            'eigcut+': _linalg.EigCutFullRank,
            'eigcut-': _linalg.EigCutLowRank,
            'svdcut+': _linalg.SVDCutFullRank,
            'svdcut-': _linalg.SVDCutLowRank,
            'lanczos': _linalg.Lanczos,
            'lobpcg': _linalg.LOBPCG,
            'chol'   : _linalg.Chol,
        }[solver]
    
    @classmethod
    def decompose(cls, posdefmatrix, solver='eigcut+', **kw):
        """
        Decompose a nonnegative definite matrix.
        
        The decomposition can be used to calculate linear algebra expressions
        where the (pseudo)inverse of the matrix appears.
        
        Parameters
        ----------
        posdefmatrix : array
            A nonnegative definite nonempty symmetric square matrix. If the
            array is not square, it must have a shape of the kind (k, n, m,
            ..., k, n, m, ...) and is reshaped to (k * n * m * ..., k * n * m *
            ...).
        solver : str
            Algorithm used to decompose the matrix.

            'eigcut+' (default)
                Promote small (or negative) eigenvalues to a minimum value.
            'eigcut-'
                Remove small (or negative) eigenvalues.
            'svdcut+'
                Promote small eigenvalues to a minimum value, keeping their
                sign.
            'svdcut-'
                Remove small eigenvalues.
            'lanczos'
                Reduce the rank of the matrix. The complexity is O(n^2 r) where
                `n` is the matrix size and `r` the required rank, while the
                other algorithms are O(n^3). Slow for small sizes.
            'lobpcg'
                Like 'lanczos' but using the LOBPCG algorithm, faster but less
                accurate.
            'chol'
                Cholesky decomposition after regularizing the matrix with a
                Gershgorin estimate of the maximum eigenvalue. The fastest of
                the O(n^3) algorithms.
        **kw :
            Additional options.

            epsrel, epsabs : positive float or 'auto'
                For solvers 'eigcut+', 'eigcut-', 'svdcut+', 'svdcut-', 'chol'.
                Specify the threshold for considering small the eigenvalues:
                
                    eps = epsrel * maximum_eigenvalue + epsabs
                
                epsrel='auto' sets epsrel = matrix_size * float_epsilon, 
                while epsabs='auto' sets epsabs = float_epsilon. Default is
                epsrel='auto', epsabs=0.
            rank : positive integer
                For the 'lanczos' and 'lobpcg' solvers, the target rank. It
                should be much smaller than the matrix size for the method to
                be convenient.
            direct_autodiff : bool
                If True, let JAX compute derivatives tracing through the code
                instead of using custom derivatives for the various operations.
                Default False.
            stop_hessian : bool
                If True, when computing second derivatives, pretend that the
                matrix has zero second derivatives w.r.t. the inputs. Does not
                work in reverse mode. Default False.
        
        Returns
        -------
        decomp : Decomposition
            An object representing the decomposition of the matrix. The
            available methods and properties are (K being the matrix):
        
            n
                The size of the matrix.
            matrix()
                Return K.
            inv()
                Compute K⁺.
            solve(b)
                Compute K⁺b.
            quad(b[, c])
                Compute b'K⁺b or b'K⁺c.
            diagquad(b)
                Compute the diagonal of b'K⁺b.
            logdet()
                Compute log(det(K)).
            tracesolve(b)
                Compute tr(K⁺b).
            correlate(b[, transpose=True])
                Compute Ab or A'b such that K = AA', with A n x m.
            decorrelate(b[, transpose=True])
                Compute A⁺b or A⁺'b with A as above.
            m
                The inner size of A.
            eps
                The threshold below which eigenvalues are not calculable.
        
        Notes
        -----
        The decomposition operations are JAX-traceable and differentiable (with
        some exceptions). Since intermediate values are stored in the
        decomposition object, it is necessary to trace a function that includes
        both the call to `decompose` and to one of the methods listed above.
        The decomposition object can be passed as input/output in jax-traceable
        functions.
        
        """
        m = jnp.asarray(posdefmatrix)
        assert m.size > 0
        assert m.ndim % 2 == 0
        half = m.ndim // 2
        head = m.shape[:half]
        tail = m.shape[half:]
        assert head == tail
        n = numpy.prod(head, dtype=int)
        m = m.reshape(n, n)
        decompcls = cls._getdecomp(solver)
        return decompcls(m, **kw)
        
        # TODO extend the interface to use composite decompositions
        # TODO accept a bufferdict for covariance matrix



# lsqfitgp/_array.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

import numpy
from numpy.lib import recfunctions
import jax
from jax import numpy as jnp
from jax import tree_util

__all__ = [
    'StructuredArray',
    'broadcast',
    'broadcast_to',
    'broadcast_arrays',
    'asarray',
    'unstructured_to_structured',
]

# TODO use register_pytree_with_keys
@tree_util.register_pytree_node_class
class StructuredArray:
    """
    JAX-friendly imitation of a numpy structured array.
    
    It behaves like a read-only numpy structured array, and you can create
    a copy with a modified field with a jax-like syntax.
    
    Examples
    --------
    >>> a = numpy.empty(3, dtype=[('f', float), ('g', float)])
    >>> a = StructuredArray(a)
    >>> a = a.at['f'].set(numpy.arange(3))
    is equivalent to
    >>> a['f'] = numpy.arange(3)
        
    Parameters
    ----------
    array : numpy array, StructuredArray
        A structured array. An array qualifies as structured if
        ``array.dtype.names is not None``.
    
    Notes
    -----
    The StructuredArray is a readonly view on the input array. When you
    change the content of a field of the StructuredArray, however, the
    reference to the original array for that field is lost.
    
    """
    
    @classmethod
    def _readonlyview_wrapifstructured(cls, x):
        if x.dtype.names is not None:
            x = cls(x)
        if isinstance(x, numpy.ndarray):
            x = x.view()
            x.flags.writeable = False
            # jax arrays and StructuredArrays are already readonly
        return x

    @classmethod
    def _array(cls, s, t, d, *, check=True):
        """
        Create a new StructuredArray.

        Parameters
        ----------
        s : tuple or None
            The shape. If None, it is determined automatically from the arrays.
        t : dtype or None
            The dtype of the array. If None, it is determined automatically
            (before the shape).
        d : dict str -> array
            The _dict of the array. The arrays, if structured, must be already
            StructuredArrays. The order of the keys must match the order of the
            fields.
        check : bool
            If True (default), check the passed values are consistent.

        Return
        ------
        out : StructuredArray
            A new StructuredArray object.
        """

        if t is None:
            # infer the data type from the arrays in the dictionary
            ndim = min((x.ndim for x in d.values()), default=None)
            t = numpy.dtype([
                (name, x.dtype, x.shape[ndim:])
                for name, x in d.items()
            ])
            # TODO infer the least common head shape instead of counting dims

        if s is None:
            # infer the shape from the arrays in the dictionary
            assert d, 'can not infer array shape with no fields'
            f = t.names[0]
            a = d[f]
            s = a.shape[:a.ndim - t[0].ndim]
        
        if check:
            assert len(t) == len(d)
            assert t.names == tuple(d.keys())
            assert all(
                x.dtype == t[f].base and x.ndim >= t[f].ndim
                for f, x in d.items()
            )
            shapes = [
                x.shape[:x.ndim - t[f].ndim]
                for f, x in d.items()
            ]
            assert all(s == s1 for s1 in shapes)
        
        out = super().__new__(cls)
        out.shape = s
        out.dtype = t
        out._dict = d
        
        return out
    
    @property
    def size(self):
        return numpy.prod(self.shape, dtype=int)
    
    @property
    def ndim(self):
        return len(self.shape)
    
    def __new__(cls, array):
        if isinstance(array, cls):
            return array
        d = {
            name: cls._readonlyview_wrapifstructured(array[name])
            for name in array.dtype.names
        }
        return cls._array(array.shape, array.dtype, d)
        
    @classmethod
    def from_dataframe(cls, df):
        """
        Make a Structured array from a DataFrame. Data is not copied.
        """
        d = {
            col: cls._readonlyview_wrapifstructured(df[col].to_numpy())
            for col in df.columns
        }
        dtype = numpy.dtype([(name, value.dtype) for name, value in d.items()])
        return cls._array((len(df),), dtype, d)
        # TODO support polars structured dtypes
    
    def __getitem__(self, key):
        if isinstance(key, str):
            return self._dict[key]
        elif isinstance(key, list) and key and all(isinstance(k, str) for k in key):
            d = {
                name: self._dict[name]
                for name in key
            }
            return self._array(self.shape, self.dtype[key], d)
        else:
            d = {
                name: x[
                    (key if isinstance(key, tuple) else (key,))
                    + (slice(None),) * self.dtype[name].ndim
                ]
                for name, x in self._dict.items()
            }
            shape = jax.eval_shape(lambda: jnp.empty(self.shape)[key]).shape
            return self._array(shape, self.dtype, d)
    
    @property
    def at(self):
        return self._Getter(self)
    
    class _Getter:
        
        def __init__(self, array):
            self.array = array
        
        def __getitem__(self, key):
            if key not in self.array.dtype.names:
                raise KeyError(key)
            return self.Setter(self.array, key)
        
        class Setter:
            
            def __init__(self, array, key, parent=None):
                self.array = array
                self.key = key
                self.parent = parent

            def __getitem__(self, subkey):
                if subkey not in self.array.dtype[self.key].names:
                    raise KeyError(subkey)
                return self.__class__(self.array[self.key], subkey, self)
            
            def set(self, val):
                assert isinstance(val, (numpy.ndarray, jnp.ndarray, StructuredArray))
                prev = self.array._dict[self.key]
                # TODO support casting and broadcasting
                assert prev.dtype == val.dtype
                assert prev.shape == val.shape
                d = dict(self.array._dict)
                d[self.key] = self.array._readonlyview_wrapifstructured(val)
                out = self.array._array(self.array.shape, self.array.dtype, d)
                if self.parent:
                    return self.parent.set(out)
                else:
                    return out
    
    def reshape(self, *shape):
        """
        Reshape the array without changing its contents. See
        numpy.ndarray.reshape.
        """
        if len(shape) == 1 and hasattr(shape[0], '__len__'):
            shape = shape[0]
        shape = tuple(shape)
        d = {
            name: x.reshape(shape + self.dtype[name].shape)
            for name, x in self._dict.items()
        }
        shape = numpy.empty(self.shape, []).reshape(shape).shape
        return self._array(shape, self.dtype, d)

    def squeeze(self, axis=None):
        """
        Remove axes of length 1. See numpy.ndarray.squeeze.
        """
        if axis is None:
            axis = tuple(i for i, size in enumerate(self.shape) if size == 1)
        if not hasattr(axis, '__len__'):
            axis = (axis,)
        assert all(self.shape[i] == 1 for i in axis)
        newshape = [size for i, size in enumerate(self.shape) if i not in axis]
        return self.reshape(newshape)

    def astype(self, dtype):
        if dtype != self.dtype:
            raise NotImplementedError
        return self
    
    def broadcast_to(self, shape, **kw):
        """
        Return a view of the array broadcasted to another shape. See
        numpy.broadcast_to.
        """
        # raises if not broadcastable
        numpy.broadcast_to(numpy.empty(self.shape, []), shape, **kw)
        d = {
            name: broadcast_to(x, shape + self.dtype[name].shape, **kw)
            for name, x in self._dict.items()
        }
        return self._array(shape, self.dtype, d)
    
    # TODO implement flatten_with_keys
    def tree_flatten(self):
        """ JAX PyTree encoder. See `jax.tree_util.tree_flatten`. """
        children = tuple(self._dict[key] for key in self.dtype.names)
        aux = dict(shape=self.shape, dtype=self.dtype)
        return children, aux
    
    @classmethod
    def tree_unflatten(cls, aux, children):
        """ JAX PyTree decoder. See `jax.tree_util.tree_unflatten`. """

        # if there are no fields, keep original shape
        if not children:
            return cls._array(aux['shape'], aux['dtype'], {})
        
        # convert children to arrays because tree_util.tree_flatten unpacks 0d
        # arrays
        children = list(map(asarray, children))
        
        # if possible, keep original dtype shapes
        oldtype = aux['dtype']
        compatible_tail_shapes = all(
            x.shape[max(0, x.ndim - oldtype[i].ndim):] == oldtype[i].shape
            for i, x in enumerate(children)
        )
        head_shapes = [
            x.shape[:max(0, x.ndim - oldtype[i].ndim)]
            for i, x in enumerate(children)
        ]
        compatible_head_shapes = all(head_shapes[0] == s for s in head_shapes)
        if compatible_tail_shapes and compatible_head_shapes:
            dtype = numpy.dtype([
                (oldtype.names[i], x.dtype, oldtype[i].shape)
                for i, x in enumerate(children)
            ])
        else:
            dtype = None

        d = dict(zip(oldtype.names, children))

        return cls._array(None, dtype, d)

        # TODO this breaks jax.jit(...).lower(...).compile()(...) because
        # apparently `lower` saves the pytree def after a step of dummyfication,
        # so the shape and dtype bases of the StructuredArray are () and object.
        # JAX expects pytrees to have a structure which does not depend on what
        # they store. => Quick hack: preserve the shape and dtype
        # unconditionally, i.e., tree_unflatten can produce malformed
        # StructuredArrays. The dictionary will contain whatever JAX puts into
        # it. => Quicker hack: it seems to me that jax always uses None as
        # dummy, so I could detect if all childrens are None or StructuredArray.

    def __repr__(self):
        # code from gvar https://github.com/gplepage/gvar
        # bufferdict.pyx:BufferDict:__str__
        out = 'StructuredArray({'

        listrepr = [(repr(k), repr(v)) for k, v in self._dict.items()]
        newlinemode = any('\n' in rv for _, rv in listrepr)
        
        for rk, rv in listrepr:
            if not newlinemode:
                out += '{}: {}, '.format(rk, rv)
            elif '\n' in rv:
                rv = rv.replace('\n', '\n    ')
                out += '\n    {}:\n    {},'.format(rk, rv)
            else:
                out += '\n    {}: {},'.format(rk, rv)
                
        if out.endswith(', '):
            out = out[:-2]
        elif newlinemode:
            out += '\n'
        out += '})'
        
        return out

        # TODO try simply using the __repr__ of self._dict
    
    def __array__(self):
        array = numpy.empty(self.shape, self.dtype)
        self._copy_into_array(array)
        return array
    
    def _copy_into_array(self, dest):
        assert self.dtype == dest.dtype
        assert self.shape == dest.shape
        for name, src in self._dict.items():
            if isinstance(src, StructuredArray):
                src._copy_into_array(dest[name])
            else:
                dest[name][...] = src

    def __array_function__(self, func, types, args, kwargs):
        if func not in self._handled_functions:
            return NotImplemented
        return self._handled_functions[func](*args, **kwargs)
    
    _handled_functions = {}
    
    @classmethod
    def _implements(cls, np_function):
        """ Register an __array_function__ implementation """
        def decorator(func):
            cls._handled_functions[np_function] = func
            func.__doc__ = textwrap.dedent(f"""\
            Implementation of {np_function.__module__}.{np_function.__name__}
            for lgp.StructuredArray.
            """) + textwrap.dedent(np_function.__doc__)
            return func
        return decorator

@StructuredArray._implements(numpy.broadcast_to)
def broadcast_to(x, shape, **kw):
    """
    Version of numpy.broadcast_to that works with StructuredArray and JAX
    arrays.
    """
    if isinstance(x, StructuredArray):
        return x.broadcast_to(shape, **kw)
    elif isinstance(x, jnp.ndarray):
        return jnp.broadcast_to(x, shape, **kw)
    else:
        return numpy.broadcast_to(x, shape, **kw)

@StructuredArray._implements(numpy.broadcast_arrays)
def broadcast_arrays(*arrays, **kw):
    """
    Version of numpy.broadcast_arrays that works with StructuredArray and JAX
    arrays.
    """
    shapes = [a.shape for a in arrays]
    shape = numpy.broadcast_shapes(*shapes)
    return [broadcast_to(a, shape, **kw) for a in arrays]
    # numpy.broadcast_arrays returns a list, not a tuple

class broadcast:
    """
    Version of numpy.broadcast that works with StructuredArray.
    """
    
    # not handled by __array_function__
    
    def __init__(self, *arrays):
        self.shape = numpy.broadcast_shapes(*(a.shape for a in arrays))

def asarray(x, dtype=None):
    """
    Version of numpy.asarray that works with StructuredArray and JAX arrays.
    If x is not an array already, returns a JAX array if possible.
    """
    if isinstance(x, (StructuredArray, jnp.ndarray, numpy.ndarray)):
        return x if dtype is None else x.astype(dtype)
    if x is None:
        return numpy.asarray(x, dtype)
        # partial workaround for jax issue #14506, None would be interpreted as
        # nan by jax
    try:
        return jnp.asarray(x, dtype)
    except (TypeError, ValueError):
        return numpy.asarray(x, dtype)

def _asarray_jaxifpossible(x):
    x = asarray(x)
    if x.dtype.names:
        return tree_util.tree_map(_asarray_jaxifpossible, StructuredArray(x))
    if isinstance(x, numpy.ndarray):
        try:
            return jnp.asarray(x)
        except (TypeError, ValueError):
            pass
    return x

@StructuredArray._implements(numpy.squeeze)
def _squeeze(a, axis=None):
    return a.squeeze(axis)

@StructuredArray._implements(numpy.ix_)
def _ix(*args):
    args = tuple(map(asarray, args))
    assert all(x.ndim == 1 for x in args)
    n = len(args)
    return tuple(
        x.reshape((1,) * i + (-1,) + (1,) * (n - i - 1))
        for i, x in enumerate(args)
    )

def unstructured_to_structured(arr,
    dtype=None,
    names=None,
    align=False, # TODO maybe align is totally inapplicable even with numpy arrays? What does it mean?
    copy=False,
    casting='unsafe'):
    """ Like numpy.lib.recfunctions.unstructured_to_structured, but outputs a
    StructuredArray. """
    arr = asarray(arr)
    if not arr.ndim:
        raise ValueError('arr must have at least one dimension')
    mockup = numpy.empty((0,) + arr.shape[-1:], arr.dtype)
    dummy = recfunctions.unstructured_to_structured(mockup,
        dtype=dtype, names=names, align=align, copy=copy, casting=casting)
    out, length = _unstructured_to_structured_recursive(0, (), arr, dummy.dtype, copy, casting)
    assert length == arr.shape[-1]
    return out

def _unstructured_to_structured_recursive(idx, shape, arr, dtype, copy, casting, *strides):
    arrays = {}
    for i, name in enumerate(dtype.names):
        base = dtype[i].base
        subshape = shape + dtype[i].shape
        size = numpy.prod(dtype[i].shape, dtype=int)
        stride = _nd(base)
        substrides = strides + ((size, stride),)
        if base.names is not None:
            y, newidx = _unstructured_to_structured_recursive(idx, subshape, arr, base, copy, casting, *substrides)
            shift = newidx - idx
            assert shift == stride
            idx += size * stride
        else:
            assert stride == 1
            if all(size == 1 for size, _ in strides):
                indices = numpy.s_[idx:idx + size]
                srcsize = size
            else:
                indices = sum((
                    stride * numpy.arange(size)[numpy.s_[:,] + (None,) * i]
                    for i, (size, stride) in enumerate(reversed(substrides))
                ), start=idx)
                indices = indices.reshape(-1)
                srcsize = indices.size
            key = numpy.s_[..., indices]
            x = arr[key]
            x = x.reshape(arr.shape[:-1] + subshape)
            if isinstance(x, jnp.ndarray):
                y = x.astype(base)
            else:
                y = x.astype(base, copy=copy, casting=casting)
            idx += size
        arrays[name] = y
    return StructuredArray._array(arr.shape[:-1] + shape, dtype, arrays), idx

@StructuredArray._implements(recfunctions.structured_to_unstructured)
def _structured_to_unstructured(arr, dtype=None, casting='unsafe'):
    mockup = numpy.empty(0, arr.dtype)
    dummy = recfunctions.structured_to_unstructured(mockup, dtype=dtype, casting=casting)
    args = (arr.shape + dummy.shape[-1:], dummy.dtype)
    try:
        out = jnp.empty(*args)
    except TypeError:
        out = numpy.empty(*args)
        # TODO can I make out column-major w.r.t. only the last column?
    out, length = _structured_to_unstructured_recursive(0, arr, out)
    assert length == dummy.shape[-1]
    return out

def _nd(dtype):
    """ Count the number of scalars in a dtype """
    base = dtype.base
    shape = dtype.shape
    size = numpy.prod(shape, dtype=int)
    if base.names is None:
        return size
    else:
        return size * sum(_nd(base[name]) for name in base.names)

def _structured_to_unstructured_recursive(idx, arr, out, *strides):
    dtype = arr.dtype
    for i, name in enumerate(dtype.names):
        subarr = arr[name]
        base = dtype[i].base
        size = numpy.prod(dtype[i].shape, dtype=int)
        stride = _nd(base)
        substrides = strides + ((size, stride),)
        if base.names is not None:
            out, newidx = _structured_to_unstructured_recursive(idx, subarr, out, *substrides)
            shift = newidx - idx
            assert shift == stride
            idx += size * stride
        else:
            assert stride == 1
            if all(size == 1 for size, _ in strides):
                indices = numpy.s_[idx:idx + size]
                srcsize = size
            else:
                indices = sum((
                    stride * numpy.arange(size)[numpy.s_[:,] + (None,) * i]
                    for i, (size, stride) in enumerate(reversed(substrides))
                ), start=idx)
                indices = indices.reshape(-1)
                srcsize = indices.size
            key = numpy.s_[..., indices]
            src = subarr.reshape(out.shape[:-1] + (srcsize,))
            if hasattr(out, 'at'):
                out = out.at[key].set(src)
            else:
                out[key] = src
            idx += size
    return out, idx

@StructuredArray._implements(numpy.empty_like)
def _empty_like(prototype, dtype=None, *, shape=None):
    shape = prototype.shape if shape is None else tuple(shape)
    dtype = prototype.dtype if dtype is None else numpy.dtype(dtype)
    return _empty(shape, dtype)

@StructuredArray._implements(numpy.empty)
def _empty(shape, dtype=float):
    arrays = {}
    for i, name in enumerate(dtype.names):
        dtbase = dtype[i].base
        dtshape = shape + dtype[i].shape
        if dtbase.names is not None:
            y = _empty(dtshape, dtbase)
        else:
            try:
                y = jnp.empty(dtshape, dtbase)
            except TypeError:
                y = numpy.empty(dtshape, dtbase)
        arrays[name] = y
    return StructuredArray._array(shape, dtype, arrays)



"""Fit of parton distributions functions (PDFs)

Version of pdf6 to do a test with JAX"""

from jax.config import config
config.update("jax_enable_x64", True)

import lsqfitgp as lgp
import jax
from jax import numpy as jnp
from autograd import numpy as anp
import numpy as np
from matplotlib import pyplot as plt
import gvar
import lsqfit

np.random.seed(20220416)

gvar._bufferdict.numpy = anp

#### SETTINGS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

indices = dict(
    # quark, antiquark
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

pid  = flavor['f0']
name = flavor['f1']

nflav  = len(flavor)

# linear data
nx        = 30 # number of PDF points used for the transformation
ndata     = 10 # number of datapoints
rankmcov  =  9 # rank of the covariance matrix of the theory error

# quadratic data
nx2       = 30 # must be <= nx
ndata2    = 10
rankmcov2 =  9

#### MODEL ####
# for each PDF:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='pid')

# grid of points to which we apply the transformation
xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

# linear map PDF(xdata) -> data
Mcomps = np.random.randn(rankmcov, ndata, nflav, nx)
Mcomps /= np.sqrt(Mcomps.size / ndata)
Mparams = gvar.gvar(np.random.randn(rankmcov), np.full(rankmcov, 0.1))
M = lambda params: jnp.tensordot(params, Mcomps, 1)

# quadratic map PDF(xdata) -> data2
M2comps = np.random.randn(rankmcov2, ndata2, nflav, nx2, nx2)
M2comps /= 2 * np.sqrt(M2comps.size / ndata2)
M2comps = (M2comps + np.swapaxes(M2comps, -1, -2)) / 2
M2params = gvar.gvar(np.random.randn(rankmcov2), np.full(rankmcov2, 0.1))
M2 = lambda params: jnp.tensordot(params, M2comps, 1)

# endpoints of the integral for each PDF
xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

# matrix to subtract the endpoints
suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### GP OBJECT ####

gp = lgp.GP()

gp.addproc(kernel, 'h')
gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
gp.addproctransf({
    'primitive': lambda x: x['x'],
    'h'        : -1,
}, 'primitive of xf(x)')

gp.addx(xdata, 'xdata', proc='f')

# linear data (used for warmup fit)
gp.addtransf({'xdata': M(gvar.mean(Mparams))}, 'data', axes=2)

# total momentum rule
gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)

# quark sum rules
qdiff = np.array([1, -1])[:, None] # vector to subtract two quarks
for quark in 'ducs':
    idx = indices[quark] # [quark, antiquark] indices
    label = f'{quark}{quark}bar' # the one appearing in `constraints`
    xlabel = f'x{label}'
    gp.addx(xinteg[idx], xlabel, proc='primitive')
    gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)
    
#### NONLINEAR FUNCTION ####

def fcn(params):
    
    xdata = params['xdata']
    Mparams = params['Mparams']
    M2params = params['M2params']
    
    data = jnp.tensordot(M(Mparams), xdata, 2)
    
    # data2 = np.einsum('dfxy,fx,fy->d', M2, xdata, xdata)
    # np.einsum does not work with gvar
    xdata2 = xdata[:, None, :nx2] * xdata[:, :nx2, None]
    data2 = jnp.tensordot(M2(M2params), xdata2, 3)
    
    return jnp.concatenate([data, data2])

prior = gp.predfromdata(constraints, ['xdata'])
prior['Mparams'] = Mparams
prior['M2params'] = M2params

gvar._bufferdict.numpy = jnp

inp = gvar.mean(prior).buf

hackfcn = lambda x: fcn(gvar.BufferDict(prior, buf=x))
jacf = jax.jit(jax.jacfwd(hackfcn))
jacr = jax.jit(jax.jacrev(hackfcn))

jacf(inp)
jacr(inp)



import jax
from jax.interpreters import ad
from jax import custom_transpose
from jax import test_util

@custom_transpose.custom_transpose
def f(x, y):
    return x * y

@f.def_transpose
def ft(ct, x, y):
    assert ad.is_undefined_primal(x) ^ ad.is_undefined_primal(y)
    if ad.is_undefined_primal(x):
        return ct * y, None
    else:
        return None, ct * x

@jax.custom_jvp
def g(x, y, z):
    return f(x, y) + z

@g.defjvp
def g_jvp(primals, tangents):
    x, y, z = primals
    xdot, ydot, zdot = tangents
    primal = g(*primals)
    tgx = f(xdot, y)
    tgy = f(x, ydot)
    return primal, tgx + tgy + zdot

test_util.check_grads(g, (1., 2., 3.), order=2, modes='fwd')



import numpy as np
from matplotlib import pyplot as plt

import lsqfitgp as lgp

fig, ax = plt.subplots(num='zeta4', clear=True)

x = np.linspace(0, 2, 10001)
for s in [1.1, 1.5, 2, 3, 4]:
    y = lgp._patch_jax.periodic_zeta_real(x, s) / lgp._patch_jax.zeta(s)
    ax.plot(x, y, label=f's={s}')

ax.legend()

fig.show()



import math
import re
import gvar

def exponent(x):
    return int(math.floor(math.log10(abs(x))))

def int_mantissa(x, n, e):
    return round(x * 10 ** (n - 1 - e))

def naive_ndigits(x, n):
    log10x = math.log10(abs(x))
    n_int = int(math.floor(n))
    n_frac = n - n_int
    log10x_frac = log10x - math.floor(log10x)
    return n_int + (log10x_frac < n_frac)

def ndigits(x, n):
    ndig = naive_ndigits(x, n)
    xexp = exponent(x)
    rounded_x = int_mantissa(x, ndig, xexp) * 10 ** xexp
    if rounded_x > x:
        rounded_ndig = naive_ndigits(rounded_x, n)
        if rounded_ndig > ndig:
            x = rounded_x
            ndig = rounded_ndig
    return x, ndig

def mantissa(x, n, e):
    m = int_mantissa(x, n, e)
    s = str(abs(int(m)))
    assert len(s) == n or len(s) == n + 1 or (m == 0 and n < 0)
    if n >= 1 and len(s) == n + 1:
        e = e + 1
        s = s[:-1]
    return s, e

def insert_dot(s, n, e, addzeros=True):
    e = e + len(s) - n
    n = len(s)
    if e >= n - 1:
        s = s + '0' * (e - n + 1)
    elif e >= 0:
        s = s[:1 + e] + '.' + s[1 + e:]
    elif e <= -1 and addzeros:
        s = '0' * -e + s
        s = s[:1] + '.' + s[1:]
    return s

def tostring(x):
    return '0' if x == 0 else f'{x:#.6g}'

def uformat(mu, s, errdig=2, sep=None, *, shareexp=True, outersign=False, uniexp=False, minnegexp=6, minposexp=4, padzeros=False, possign=False):
    """
    Format a number with uncertainty.
    
    Parameters
    ----------
    mu : number
        The central value.
    s : number
        The error.
    errdig : number
        The number of digits of the error to be shown. Must be >= 1. It can be
        a noninteger, in which case the number of digits switches between the
        lower nearest integer to the upper nearest integer as the first decimal
        digit (after rounding) crosses 10 raised to the fractional part of
        `errdig`. Default 1.5.
    sep : None or str
        The separator put between the central value and the error. Eventual
        spaces must be included. If None, put the error between parentheses,
        sharing decimal places/exponential notation with the central value.
        Default None.
    shareexp : bool
        Applies if sep is not None. When using exponential notation, whether to
        share the exponent between central value and error with outer
        parentheses. Default True.
    outersign : bool
        Applied when sep is not None and shareexp is True. Whether to put the
        sign outside or within the parentheses. Default False
    uniexp : bool
        When using exponential notation, whether to use unicode characters
        instead of the standard ASCII notation. Default False.
    minnegexp : int
        The number of places after the comma at which the notation switches
        to exponential notation. Default 4. The number of places from the
        greater between central value and error is considered.
    minposexp : int
        The power of ten of the least significant digit at which exponential
        notation is used. Default 0. Setting higher values may force padding
        the error with zeros, depending on `errdig`.
    padzeros : bool
        Whether to pad with zeros when not using exponential notation due to
        `minposexp` even if the least significant digits is not on the units.
        Default False, i.e., more digits than those specified
    possign : bool
        Whether to put a `+` before the central value when it is positive.
        Default False.
    """
    if errdig < 1:
        raise ValueError('errdig < 1')
    if not math.isfinite(mu) or not math.isfinite(s) or s <= 0:
        if sep is None:
            return f'{tostring(mu)}({tostring(s)})'
        else:
            return f'{tostring(mu)}{sep}{tostring(s)}'
    
    s, sndig = ndigits(s, errdig)
    sexp = exponent(s)
    muexp = exponent(mu) if mu != 0 else sexp - sndig - 1
    smant, sexp = mantissa(s, sndig, sexp)
    mundig = sndig + muexp - sexp
    mumant, muexp = mantissa(mu, mundig, muexp)
    musign = '-' if mu < 0 else '+' if possign else ''
    
    if mundig >= sndig:
        use_exp = muexp >= mundig + minposexp or muexp <= -minnegexp
        base_exp = muexp
    else:
        use_exp = sexp >= sndig + minposexp or sexp <= -minnegexp
        base_exp = sexp
    
    if use_exp:
        mumant = insert_dot(mumant, mundig, muexp - base_exp)
        smant = insert_dot(smant, sndig, sexp - base_exp, sep is not None)
    elif base_exp >= max(mundig, sndig) and not padzeros:
        mumant = str(abs(round(mu)))
        smant = str(abs(round(s)))
    else:
        mumant = insert_dot(mumant, mundig, muexp)
        smant = insert_dot(smant, sndig, sexp, sep is not None)
    
    if not outersign:
        mumant = musign + mumant
    
    if use_exp:
        if uniexp:
            asc = '0123456789+-'
            uni = '⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻'
            table = str.maketrans(asc, uni)
            exp = str(base_exp).translate(table)
            suffix = '×10' + exp
        else:
            suffix = f'e{base_exp:+}'
        if sep is None:
            r = mumant + '(' + smant + ')' + suffix
        elif shareexp:
            r = '(' + mumant + sep + smant + ')' + suffix
        else:
            r = mumant + suffix + sep + smant + suffix
    elif sep is None:
        r = mumant + '(' + smant + ')'
    else:
        r = mumant + sep + smant
    
    if outersign:
        r = musign + r
    
    return r

def check(n, s, string, *args, **kw):
    defaults = dict(minnegexp=2, minposexp=0)
    defaults.update(kw)
    f = uformat(n, s, *args, **defaults)
    if f != string:
        raise RuntimeError(f'{f!r} != {string!r}')

def allchecks():
    check(1, 0.2, "1.00 pm 0.20", 1.5, " pm ")
    check(1, 0.3, "1.00 pm 0.30", 1.5, " pm ")
    check(1, 0.31, "1.00 pm 0.31", 1.5, " pm ")
    check(1, 0.32, "1.0 pm 0.3", 1.5, " pm ")
    check(-1, 0.34, "-1.00 pm 0.34", 2, " pm ")
    check(0, 0, "0 pm 0", 2, " pm ")
    check(123456, 0, "123456. pm 0", 2, " pm ")
    check(12345.6, 0, "12345.6 pm 0", 2, " pm ")
    check(12345.67, 0, "12345.7 pm 0", 2, " pm ")
    check(1e8, 0, "1.00000e+08 pm 0", 2, " pm ")
    check(1e-2, 0, "0.0100000 pm 0", 2, " pm ")
    check(1e-1, 0, "0.100000 pm 0", 2, " pm ")
    check(12345.99, 0, "12346.0 pm 0", 2, " pm ")
    check(0, 0.001, "(0.0 pm 1.0)e-3", 2, " pm ")
    check(0, 0.01, "(0.0 pm 1.0)e-2", 2, " pm ")
    check(0, 0.1, "0.00 pm 0.10", 2, " pm ")
    check(0, 1, "0.0 pm 1.0", 2, " pm ")
    check(0, 10, "0 pm 10", 2, " pm ")
    check(0, 100, "(0.0 pm 1.0)e+2", 2, " pm ")
    check(0, 1000, "(0.0 pm 1.0)e+3", 2, " pm ")
    check(0, 0.0196, "(0.0 pm 2.0)e-2", 2, " pm ")
    check(0, 0.196, "0.00 pm 0.20", 2, " pm ")
    check(0, 1.96, "0.0 pm 2.0", 2, " pm ")
    check(0, 19.6, "0 pm 20", 2, " pm ")
    check(0, 196, "(0.0 pm 2.0)e+2", 2, " pm ")
    check(0, 0.00996, "(0.0 pm 1.0)e-2", 2, " pm ")
    check(0, 0.0996, "0.00 pm 0.10", 2, " pm ")
    check(0, 0.996, "0.0 pm 1.0", 2, " pm ")
    check(0, 9.96, "0 pm 10", 2, " pm ")
    check(0, 99.6, "(0.0 pm 1.0)e+2", 2, " pm ")
    check(0.025, 3, "0.0 pm 3.0", 2, " pm ")
    check(0.0251, 0.3, "0.03 pm 0.30", 2, " pm ")
    check(0.025, 0.03, "(2.5 pm 3.0)e-2", 2, " pm ")
    check(0.025, 0.003, "(2.50 pm 0.30)e-2", 2, " pm ")
    check(0.0025, 0.003, "(2.5 pm 3.0)e-3", 2, " pm ")
    check(0.251, 3, "0.3 pm 3.0", 2, " pm ")
    check(2.5, 3, "2.5 pm 3.0", 2, " pm ")
    check(25, 3, "25.0 pm 3.0", 2, " pm ")
    check(2500, 300, "(2.50 pm 0.30)e+3", 2, " pm ")
    check(1, 0.99, "1.0 pm 1.0", 1.5, " pm ")
    check(math.inf, 1.0, "inf pm 1.00000", 2, " pm ")
    check(-math.inf, 1.0, "-inf pm 1.00000", 2, " pm ")
    check(0, math.inf, "0 pm inf", 2, " pm ")

    check(1, 0.2, "1.00(20)", 1.5, None)
    check(1, 0.3, "1.00(30)", 1.5, None)
    check(1, 0.31, "1.00(31)", 1.5, None)
    check(1, 0.32, "1.0(3)", 1.5, None)
    check(-1, 0.34, "-1.00(34)", 2, None)
    check(0, 0, "0(0)", 2, None)
    check(123456, 0, "123456.(0)", 2, None)
    check(12345.6, 0, "12345.6(0)", 2, None)
    check(12345.67, 0, "12345.7(0)", 2, None)
    check(1e8, 0, "1.00000e+08(0)", 2, None)
    check(1e-2, 0, "0.0100000(0)", 2, None)
    check(1e-1, 0, "0.100000(0)", 2, None)
    check(12345.99, 0, "12346.0(0)", 2, None)
    check(0, 0.001, "0.0(1.0)e-3", 2, None)
    check(0, 0.01, "0.0(1.0)e-2", 2, None)
    check(0, 0.1, "0.00(10)", 2, None)
    check(0, 1, "0.0(1.0)", 2, None)
    check(0, 10, "0(10)", 2, None)
    check(0, 100, "0.0(1.0)e+2", 2, None)
    check(0, 1000, "0.0(1.0)e+3", 2, None)
    check(0, 0.0196, "0.0(2.0)e-2", 2, None)
    check(0, 0.196, "0.00(20)", 2, None)
    check(0, 1.96, "0.0(2.0)", 2, None)
    check(0, 19.6, "0(20)", 2, None)
    check(0, 196, "0.0(2.0)e+2", 2, None)
    check(0, 0.00996, "0.0(1.0)e-2", 2, None)
    check(0, 0.0996, "0.00(10)", 2, None)
    check(0, 0.996, "0.0(1.0)", 2, None)
    check(0, 9.96, "0(10)", 2, None)
    check(0, 99.6, "0.0(1.0)e+2", 2, None)
    check(0.025, 3, "0.0(3.0)", 2, None)
    check(0.0251, 0.3, "0.03(30)", 2, None)
    check(0.025, 0.03, "2.5(3.0)e-2", 2, None)
    check(0.025, 0.003, "2.50(30)e-2", 2, None)
    check(0.0025, 0.003, "2.5(3.0)e-3", 2, None)
    check(0.251, 3, "0.3(3.0)", 2, None)
    check(2.5, 3, "2.5(3.0)", 2, None)
    check(25, 3, "25.0(3.0)", 2, None)
    check(2500, 300, "2.50(30)e+3", 2, None)
    check(1, 0.99, "1.0(1.0)", 1.5, None)
    check(math.inf, 1.0, "inf(1.00000)", 2, None)
    check(-math.inf, 1.0, "-inf(1.00000)", 2, None)
    check(0, math.inf, "0(inf)", 2, None)

class NewGVar(gvar._gvarcore.GVar):
    
    def __init__(self, g):
        self.g = g

    def __format__(self, spec):
        self = self.g
        if not spec:
            return str(self)
        pat = r'([-+#$]*)(\d*\.?\d*)(:\d+)?(p|s|u|U)'
        m = re.fullmatch(pat, spec)
        if not m:
            raise ValueError(f'format specification {spec!r} not understood, format is r"{pat}"')
        kw = {}
        options = m.group(1)
        kw['possign'] = '+' in options
        kw['outersign'] = '-' in options
        kw['padzeros'] = '#' in options
        kw['shareexp'] = '$' not in options
        if m.group(2):
            kw['errdig'] = float(m.group(2))
        else:
            kw['errdig'] = 1.5
        if m.group(3):
            nexp = int(m.group(3)[1:])
        else:
            nexp = 5
        kw['minposexp'] = max(0, nexp - math.floor(kw['errdig']))
        kw['minnegexp'] = nexp
        mode = m.group(4)
        kw['sep'] = dict(p=None, s=' +/- ', u=' ± ', U=' ± ')[mode]
        kw['uniexp'] = mode == 'U'
        return uformat(gvar.mean(self), gvar.sdev(self), **kw)

def gv(*args):
    return NewGVar(gvar.gvar(*args))

def compare(n, s):
    print(f'{gv(n, s)}      {gv(n, s):p}')

if __name__ == '__main__':
    allchecks()
